"""Hamiltonian systems encapsulating energy functions and their derivatives."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from mici import matrices
from mici.autodiff import autodiff_fallback, wrap_function
from mici.states import cache_in_state, cache_in_state_with_aux

if TYPE_CHECKING:
    from typing import Any

    from numpy.random import Generator

    from mici.states import ChainState
    from mici.types import (
        ArrayFunction,
        ArrayLike,
        GradientFunction,
        HessianFunction,
        JacobianFunction,
        MatrixHessianProduct,
        MatrixHessianProductFunction,
        MatrixLike,
        MatrixTressianProduct,
        MatrixTressianProductFunction,
        MetricLike,
        ScalarFunction,
        ScalarLike,
        VectorJacobianProduct,
        VectorJacobianProductFunction,
    )


class System(ABC):
    r"""Base class for Hamiltonian systems.

    The Hamiltonian function :math:`h` is assumed to have the general form

    .. math::

        h(q, p) = h_1(q) + h_2(q, p)

    where :math:`q` and :math:`p` are the position and momentum variables respectively,
    and :math:`h_1` and :math:`h_2` Hamiltonian component functions. The exact
    Hamiltonian flow for the :math:`h_1` component can be always be computed as it
    depends only on the position variable however depending on the form of :math:`h_2`
    the corresponding exact Hamiltonian flow may or may not be simulable.

    By default :math:`h_1` is assumed to correspond to the negative logarithm of an
    unnormalized density on the position variables with respect to a reference measure,
    with the corresponding distribution on the position space being the target
    distribution it is wished to draw approximate samples from.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        *,
        grad_neg_log_dens: GradientFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to a reference measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        self._neg_log_dens = wrap_function(neg_log_dens, backend)
        self._grad_neg_log_dens = autodiff_fallback(
            grad_neg_log_dens,
            neg_log_dens,
            "grad_and_value",
            "grad_neg_log_dens",
            backend,
        )

    @cache_in_state("pos")
    def neg_log_dens(self, state: ChainState) -> ScalarLike:
        """Negative logarithm of unnormalized density of target distribution.

        Args:
            state: State to compute value at.

        Returns:
            Value of computed negative log density.
        """
        return self._neg_log_dens(state.pos)

    @cache_in_state_with_aux("pos", "neg_log_dens")
    def grad_neg_log_dens(self, state: ChainState) -> ArrayLike:
        """Derivative of negative log density with respect to position.

        Args:
            state: State to compute value at.

        Returns:
            Value of `neg_log_dens(state)` derivative with respect to `state.pos`.
        """
        return self._grad_neg_log_dens(state.pos)

    def h1(self, state: ChainState) -> ScalarLike:
        """Hamiltonian component depending only on position.

        Args:
            state: State to compute value at.

        Returns:
            Value of `h1` Hamiltonian component.
        """
        return self.neg_log_dens(state)

    def dh1_dpos(self, state: ChainState) -> ScalarLike:
        """Derivative of `h1` Hamiltonian component with respect to position.

        Args:
            state: State to compute value at.

        Returns:
            Value of computed `h1` derivative.
        """
        return self.grad_neg_log_dens(state)

    def h1_flow(self, state: ChainState, dt: ScalarLike) -> None:
        """Apply exact flow map corresponding to `h1` Hamiltonian component.

        `state` argument is modified in place.

        Args:
            state: State to start flow at.
            dt: Time interval to simulate flow for.
        """
        state.mom -= dt * self.dh1_dpos(state)

    @abstractmethod
    def h2(self, state: ChainState) -> ScalarLike:
        """Hamiltonian component depending on momentum and optionally position.

        Args:
            state: State to compute value at.

        Returns:
            Value of `h2` Hamiltonian component.
        """

    @abstractmethod
    def dh2_dpos(self, state: ChainState) -> ArrayLike:
        """Derivative of `h2` Hamiltonian component with respect to position.

        Args:
            state: State to compute value at.

        Returns:
            Value of `h2(state)` derivative with respect to `state.pos`.
        """

    @abstractmethod
    def dh2_dmom(self, state: ChainState) -> ArrayLike:
        """Derivative of `h2` Hamiltonian component with respect to momentum.

        Args:
            state: State to compute value at.

        Returns:
            Value of `h2(state)` derivative with respect to `state.mom`.
        """

    def h(self, state: ChainState) -> ScalarLike:
        """Hamiltonian function for system.

        Args:
            state: State to compute value at.

        Returns:
            Value of Hamiltonian.
        """
        return self.h1(state) + self.h2(state)

    def dh_dpos(self, state: ChainState) -> ArrayLike:
        """Derivative of Hamiltonian with respect to position.

        Args:
            state: State to compute value at.

        Returns:
            Value of `h(state)` derivative with respect to `state.pos`.
        """
        return self.dh1_dpos(state) + self.dh2_dpos(state)

    def dh_dmom(self, state: ChainState) -> ArrayLike:
        """Derivative of Hamiltonian with respect to momentum.

        Args:
            state: State to compute value at.

        Returns:
            Value of `h(state)` derivative with respect to `state.mom`.
        """
        return self.dh2_dmom(state)

    @abstractmethod
    def sample_momentum(self, state: ChainState, rng: Generator) -> ArrayLike:
        """Sample a momentum from its conditional distribution given a position.

        Args:
            state: State defining position to condition on.

        Returns:
            Sampled momentum.
        """


class TractableFlowSystem(System):
    r"""Base class for Hamiltonian systems with tractable component flows.

    The Hamiltonian function :math:`h` is assumed to have the general form

    .. math::

        h(q, p) = h_1(q) + h_2(q, p)

    where :math:`q` and :math:`p` are the position and momentum variables respectively,
    and :math:`h_1` and :math:`h_2` Hamiltonian component functions. The exact
    Hamiltonian flows for both the :math:`h_1` and :math:`h_2` components are assumed to
    be tractable for subclasses of this class.

    By default :math:`h_1` is assumed to correspond to the negative logarithm of an
    unnormalized density on the position variables with respect to a reference measure,
    with the corresponding distribution on the position space being the target
    distribution it is wished to draw approximate samples from.
    """

    @abstractmethod
    def h2_flow(self, state: ChainState, dt: ScalarLike) -> None:
        """Apply exact flow map corresponding to `h2` Hamiltonian component.

        `state` argument is modified in place.

        Args:
            state: State to start flow at.
            dt: Time interval to simulate flow for.
        """


class EuclideanMetricSystem(TractableFlowSystem):
    r"""Hamiltonian system with a Euclidean metric on the position space.

    Here Euclidean metric is defined to mean a metric with a fixed positive definite
    matrix representation :math:`M`. The momentum variables are taken to be independent
    of the position variables and with a zero-mean Gaussian marginal distribution with
    covariance specified by :math:`M`, so that the :math:`h_2` Hamiltonian component is

    .. math::

        h_2(q, p) = \frac{1}{2} p^T M^{-1} p

    where :math:`q` and :math:`p` are the position and momentum variables respectively.

    The :math:`h_1` Hamiltonian component function is

    .. math::

        h_1(q) = \ell(q)

    where :math:`\ell(q)` is the negative log (unnormalized) density of the target
    distribution with respect to the Lebesgue measure.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        *,
        metric: MetricLike | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the Lebesgue measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            metric: Matrix object corresponding to matrix representation of metric on
                position space and covariance of Gaussian marginal distribution on
                momentum vector. If `None` is passed (the default), the identity matrix
                will be used. If a 1D array is passed then this is assumed to specify a
                metric with positive diagonal matrix representation and the array the
                matrix diagonal. If a 2D array is passed then this is assumed to specify
                a metric with a dense positive definite matrix representation specified
                by the array. Otherwise if the value is a subclass of
                `mici.matrices.PositiveDefiniteMatrix` it is assumed to directly specify
                the metric matrix representation.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            grad_neg_log_dens=grad_neg_log_dens,
            backend=backend,
        )
        if metric is None:
            self.metric = matrices.IdentityMatrix()
        elif isinstance(metric, np.ndarray):
            if metric.ndim == 1:
                self.metric = matrices.PositiveDiagonalMatrix(metric)
            elif metric.ndim == 2:  # noqa: PLR2004
                self.metric = matrices.DensePositiveDefiniteMatrix(metric)
            else:
                msg = (
                    "If NumPy ndarray value is used for `metric` must be either 1D "
                    "(diagonal matrix) or 2D (dense positive definite matrix)."
                )
                raise ValueError(msg)
        else:
            self.metric = metric

    @cache_in_state("mom")
    def h2(self, state: ChainState) -> ScalarLike:
        return 0.5 * state.mom @ self.dh2_dmom(state)

    @cache_in_state("mom")
    def dh2_dmom(self, state: ChainState) -> ArrayLike:
        return self.metric.inv @ state.mom

    def dh2_dpos(self, state: ChainState) -> ArrayLike:
        return np.zeros_like(state.pos)

    def dh_dpos(self, state: ChainState) -> ArrayLike:
        return self.dh1_dpos(state)

    def h2_flow(self, state: ChainState, dt: ScalarLike) -> None:
        state.pos += dt * self.dh2_dmom(state)

    def sample_momentum(self, state: ChainState, rng: Generator) -> ArrayLike:
        return self.metric.sqrt @ rng.standard_normal(state.pos.shape)


class GaussianEuclideanMetricSystem(EuclideanMetricSystem):
    r"""Euclidean Hamiltonian system with a tractable Gaussian component.

    Here Euclidean metric is defined to mean a metric with a fixed positive definite
    matrix representation :math:`M`. The momentum variables are taken to be independent
    of the position variables and with a zero-mean Gaussian marginal distribution with
    covariance specified by :math:`M`.

    Additionally the target distribution on the position variables is assumed to be
    defined by an unnormalized density with respect to the standard Gaussian measure on
    the position space (with identity covariance and zero mean), with the Hamiltonian
    component :math:`h_1` corresponding to the negative logarithm of this density rather
    than the density with respect to the Lebesgue measure on the position space, i.e.

    .. math::

        h_1(q) = \ell(q) - \frac{1}{2} q^T q

    where :math:`q` is the position and :math:`\ell(q)` is the negative log
    (unnormalized) density of the target distribution with respect to the Lebesgue
    measure at :math:`q`. The Hamiltonian  component function :math:`h_2` is then
    assumed to have the form

    .. math::

        h_2(q, p) = \frac{1}{2} q^T q + \frac{1}{2} p^T M^{-1} p

    where :math:`p` is the momentum. In this case the Hamiltonian flow due to the
    quadratic :math:`h_2` component can be solved for analytically, allowing an
    integrator to be defined using this alternative splitting of the Hamiltonian [1].

    References:
      1. Shahbaba, B., Lan, S., Johnson, W.O. and Neal, R.M., 2014. Split Hamiltonian
         Monte Carlo. Statistics and Computing, 24(3), pp.339-349.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        *,
        metric: MetricLike | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the standard Gaussian measure on the position space,
                with the corresponding distribution on the position space being the
                target distribution it is wished to draw approximate samples from.
            metric: Matrix object corresponding to matrix representation of metric on
                position space and covariance of Gaussian marginal distribution on
                momentum vector. If `None` is passed (the default), the identity matrix
                will be used. If a 1D array is passed then this is assumed to specify a
                metric with positive diagonal matrix representation and the array the
                matrix diagonal. If a 2D array is passed then this is assumed to specify
                a metric with a dense positive definite matrix representation specified
                by the array. Otherwise if the value is a subclass of
                `mici.matrices.PositiveDefiniteMatrix` it is assumed to directly specify
                the metric matrix representation.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            metric=metric,
            grad_neg_log_dens=grad_neg_log_dens,
            backend=backend,
        )

    def h2(self, state: ChainState) -> ScalarLike:
        return (
            0.5 * state.pos @ state.pos + 0.5 * state.mom @ self.metric.inv @ state.mom
        )

    @cache_in_state("mom")
    def dh2_dmom(self, state: ChainState) -> ArrayLike:
        return self.metric.inv @ state.mom

    @cache_in_state("mom")
    def dh2_dpos(self, state: ChainState) -> ArrayLike:
        return state.pos

    def h2_flow(self, state: ChainState, dt: ScalarLike) -> None:
        omega = 1.0 / self.metric.eigval**0.5
        sin_omega_dt, cos_omega_dt = np.sin(omega * dt), np.cos(omega * dt)
        eigvec_trans_pos = self.metric.eigvec.T @ state.pos
        eigvec_trans_mom = self.metric.eigvec.T @ state.mom
        state.pos = self.metric.eigvec @ (
            cos_omega_dt * eigvec_trans_pos + (sin_omega_dt * omega) * eigvec_trans_mom
        )
        state.mom = self.metric.eigvec @ (
            cos_omega_dt * eigvec_trans_mom - (sin_omega_dt / omega) * eigvec_trans_pos
        )


class ConstrainedTractableFlowSystem(TractableFlowSystem):
    r"""Base class for Hamiltonian systems subject to constraints with tractable flows.

    The (constrained) position space is assumed to be a differentiable manifold embedded
    with a :math:`Q`-dimensional ambient Euclidean space. The :math:`Q-C` dimensional
    manifold :math:`\mathcal{M}` is implicitly defined by an equation

    .. math::

        \mathcal{M} = \lbrace q \in \mathbb{R}^Q : c(q) = 0 \rbrace

    with :math:`c: \mathbb{R}^Q \to \mathbb{R}^C` the differentiable and surjective
    vector-valued *constraint function*.

    The Hamiltonian function :math:`h` is assumed to have the general form

    .. math::

        h(q, p) = h_1(q) + h_2(q, p)

    where :math:`q` and :math:`p` are the position and momentum variables respectively,
    and :math:`h_1` and :math:`h_2` Hamiltonian component functions. The exact
    *unconstrained* Hamiltonian flows for both the :math:`h_1` and :math:`h_2`
    components, respectively :math:`\Phi_1` and :math:`\Phi_2` are assumed to be
    tractable for subclasses of this class.

    The *constrained* Hamiltonian dynamics are described by the system of differential
    algebraic equations

    .. math::

        \dot{q} = \nabla_2 h(q, p), \quad
        \dot{p} = -\nabla_1 h(q, p) - \partial c(q)^T\lambda, \quad
        c(q) = 0.

    where :math:`\lambda` is a set of Lagrange multipliers of dimension equal to number
    of constraints, :math:`C`, and which are implicitly defined by the condition that
    the constraint equation :math:`c(q) = 0` applies at all times.

    The dynamics implicitly define a set of constraints on the momentum variables,
    differentiating the constraint equation with respect to time giving that

    .. math::

        \partial c(q) \nabla_2 h(q, p) = \partial c(q) \nabla_2 h_2(q, p) = 0.

    The set of momentum variables satisfying the above for given position variables is
    termed the cotangent space of the manifold (at a position). Here we assume that the
    operation of projecting a momentum vector onto the cotangent space at a given
    position is tractable to compute.
    """

    @abstractmethod
    def constr(self, state: ChainState) -> ArrayLike:
        """Constraint function at the current position.

        Args:
            state: State to compute value at.

        Returns:
            Value of :code:`constr(state.pos)` as 1D array.
        """

    @abstractmethod
    def jacob_constr(self, state: ChainState) -> ArrayLike:
        """Jacobian of constraint function at the current position.

        Args:
            state: State to compute value at.

        Returns:
            Value of Jacobian of :code:`constr(state.pos)` as 2D array.
        """

    @abstractmethod
    def jacob_constr_inner_product(
        self,
        jacob_constr_1: MatrixLike,
        inner_product_matrix: matrices.PositiveDefiniteMatrix,
        jacob_constr_2: MatrixLike | None = None,
    ) -> MatrixLike:
        """Compute inner product of rows of constraint Jacobian matrices.

        Computes :code:`jacob_constr_1 @ inner_product_matrix @ jacob_constr_2.T`
        potentially exploiting any structure / sparsity in :code:`jacob_constr_1`,
        :code:`jacob_constr_2` and :code:`inner_product_matrix`.

        Args:
            jacob_constr_1: First constraint Jacobian in product.
            inner_product_matrix: Positive-definite matrix defining inner-product
                between rows of two constraint Jacobians.
            jacob_constr_2: Second constraint Jacobian in product. Defaults to
                :code:`jacob_constr_1` if set to :code:`None`.

        Returns:
            Object corresponding to computed inner products of the constraint Jacobian
            rows.
        """

    @abstractmethod
    def dh2_flow_dmom(
        self,
        state: ChainState,
        dt: ScalarLike,
    ) -> tuple[matrices.Matrix, matrices.Matrix]:
        """Derivatives of :py:meth:`h2_flow` flow map with respect to momentum argument.

        Args:
            state: State to evaluate derivatives of flow map at.
            dt: Time interval flow simulated for.

        Returns:
            Tuple :code:`(dpos_dmom, dmom_dmom)` with :code:`dpos_dmom` a matrix
            representing derivative (Jacobian) of position output of :py:meth:`h2_flow`
            with respect to the value of the momentum component of the initial input
            state and :code:`dmom_dmom` a matrix representing derivative (Jacobian) of
            momentum output of :py:meth:`h2_flow` with respect to the value of the
            momentum component of the initial input state.
        """

    @abstractmethod
    def project_onto_cotangent_space(
        self,
        mom: ArrayLike,
        state: ChainState,
    ) -> ArrayLike:
        """Project a momentum on to the co-tangent space at a position.

        Args:
            mom: Momentum (co-)vector as 1D array to project on to co-tangent space.
            state: State definining position on the  manifold to project in to the
                co-tangent space of.

        Returns:
            Projected momentum in the co-tangent space at :code:`state.pos`.
        """

    def sample_momentum(self, state: ChainState, rng: Generator) -> ArrayLike:
        mom = super().sample_momentum(state, rng)
        return self.project_onto_cotangent_space(mom, state)


class ConstrainedEuclideanMetricSystem(
    ConstrainedTractableFlowSystem,
    EuclideanMetricSystem,
):
    r"""Base class for Euclidean Hamiltonian systems subject to constraints.

    The (constrained) position space is assumed to be a differentiable manifold embedded
    with a :math:`Q`-dimensional ambient Euclidean space. The :math:`Q-C` dimensional
    manifold :math:`\mathcal{M}` is implicitly defined by an equation
    :math:`\mathcal{M} = \lbrace q \in \mathbb{R}^Q : c(q) = 0 \rbrace` with
    :math:`c: \mathbb{R}^Q \to \mathbb{R}^C` the *constraint function*.

    The ambient Euclidean space is assumed to be equipped with a metric with constant
    positive-definite matrix representation :math:`M` which further specifies the
    covariance of the zero-mean Gaussian distribution :math:`\mathcal{N}(0, M)` on the
    *unconstrained* momentum (co-)vector :math:`p` with corresponding :math:`h_2`
    Hamiltonian component defined as

    .. math::

        h_2(q, p) = \frac{1}{2} p^T M^{-1} p.

    The time-derivative of the constraint equation implies a further set of constraints
    on the momentum :math:`p` with :math:`\partial c(q) M^{-1} p = 0` at all time
    points, corresponding to the momentum (velocity) being in the co-tangent space
    (tangent space) to the manifold.

    The target distribution is either assumed to be directly specified with unnormalized
    density :math:`\exp(-\ell(q))` with respect to the Hausdorff measure on the manifold
    (under the metric induced from the ambient metric) with in this case the :math:`h_1`
    Hamiltonian component then simply

    .. math::

        h_1(q) = \ell(q),

    or alternatively it is assumed a prior distribution on the position :math:`q` with
    density :math:`\exp(-\ell(q))` with respect to the Lebesgue measure on the ambient
    space is specifed and the target distribution is the posterior distribution on
    :math:`q` when conditioning on the event :math:`c(q) = 0`. The negative logarithm of
    the posterior distribution density with respect to the Hausdorff measure (and so
    :math:`h_1` Hamiltonian component) is then

    .. math::

        h_1(q) =
        \ell(q) + \frac{1}{2} \log\left|\partial c(q)M^{-1}\partial c(q)^T\right|

    with an additional second *Gram matrix* determinant term to give the correct density
    with respect to the Hausdorff measure on the manifold.

    Due to the requirement to enforce the constraints on the position and momentum, a
    constraint-preserving numerical integrator needs to be used when simulating the
    Hamiltonian dynamic associated with the system, e.g.
    :py:class:`mici.integrators.ConstrainedLeapfrogIntegrator`.

    References:
      1. Lelièvre, T., Rousset, M. and Stoltz, G., 2019. Hybrid Monte Carlo methods for
         sampling probability measures on submanifolds. Numerische Mathematik, 143(2),
         pp.379-421.
      2. Graham, M.M. and Storkey, A.J., 2017. Asymptotically exact inference in
         differentiable generative models. Electronic Journal of Statistics, 11(2),
         pp.5105-5164.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        constr: ArrayFunction,
        *,
        metric: MetricLike | None = None,
        dens_wrt_hausdorff: bool = True,
        grad_neg_log_dens: GradientFunction | None = None,
        jacob_constr: JacobianFunction | None = None,
        backend: str | None = None,
    ) -> None:
        r"""
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the constrained
                position space with respect to the Hausdorff measure on the constraint
                manifold (if :code:`dens_wrt_hausdorff == True`) or alternatively the
                negative logarithm of an unnormalized probability density on the
                unconstrained (ambient) position space with respect to the Lebesgue
                measure. In the former case the target distribution it is wished to draw
                approximate samples from is assumed to be directly specified by the
                density function on the manifold. In the latter case the density
                function is instead taken to specify a prior distribution on the ambient
                space with the target distribution then corresponding to the posterior
                distribution when conditioning on the (zero Lebesgue measure) event
                :code:`constr(q) == 0` where :code:`q` is the position array. This
                target posterior distribution has support on the differentiable manifold
                implicitly defined by the constraint equation, with density with respect
                to the Hausdorff measure on the manifold corresponding to the ratio of
                the prior density (specified by :code:`neg_log_dens`) and the
                square-root of the determinant of the Gram matrix defined by
                :code:`gram(q) = jacob_constr(q) @ inv(metric) @ jacob_constr(q).T`
                where :code:`jacob_constr` is the Jacobian of the constraint function
                :code:`constr` and :code:`metric` is the matrix representation of the
                metric on the ambient space.
            constr: Function which given a position array return as a 1D array the value
                of the (vector-valued) constraint function, the zero level-set of which
                implicitly defines the manifold the dynamic is simulated on.
            metric: Matrix object corresponding to matrix representation of metric on
                *unconstrained* position space and covariance of Gaussian marginal
                distribution on *unconstrained* momentum vector. If :code:`None` is
                passed (the default), the identity matrix will be used. If a 1D array is
                passed then this is assumed to specify a metric with positive diagonal
                matrix representation and the array the matrix diagonal. If a 2D array
                is passed then this is assumed to specify a metric with a dense positive
                definite matrix representation specified by the array. Otherwise if the
                value is a :py:class:`mici.matrices.PositiveDefiniteMatrix` subclass it
                is assumed to directly specify the metric matrix representation.
            dens_wrt_hausdorff: Whether the :code:`neg_log_dens` function specifies the
                (negative logarithm) of the density of the target distribution with
                respect to the Hausdorff measure on the manifold directly (:code:`True`)
                or alternatively the negative logarithm of a density of a prior
                distriubtion on the unconstrained (ambient) position space with respect
                to the Lebesgue measure, with the target distribution then corresponding
                to the posterior distribution when conditioning on the event
                :code:`constr(pos) == 0` (:code:`False`). Note that in the former case
                the base Hausdorff measure on the manifold depends on the metric defined
                on the ambient space, with the Hausdorff measure being defined with
                respect to the metric induced on the manifold from this ambient metric.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of :code:`neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the :code:`neg_log_dens` evaluated at the
                passed position array. If :code:`None` is passed (the default) an
                automatic differentiation fallback will be used to attempt to construct
                a function to compute the derivative (and value) of :code:`neg_log_dens`
                automatically.
            jacob_constr: Function which given a position array computes the Jacobian
                (matrix / 2D array of partial derivatives) of the output of the
                constraint function :code:`c = constr(q)` with respect to the position
                array argument :code:`q`, returning the computed Jacobian as a 2D array
                :code:`jacob` with :code:`jacob[i, j] = ∂c[i] / ∂q[j]`. Optionally the
                function may instead return a 2-tuple of values with the first being the
                array corresponding to the Jacobian and the second being the value of
                :code:`constr` evaluated at the passed position array. If :code:`None`
                is passed (the default) an automatic differentiation fallback will be
                used to attempt to construct a function to compute the Jacobian (and
                value) of :code:`constr`
                automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            metric=metric,
            grad_neg_log_dens=grad_neg_log_dens,
            backend=backend,
        )
        self._constr = wrap_function(constr, backend)
        self.dens_wrt_hausdorff = dens_wrt_hausdorff
        self._jacob_constr = autodiff_fallback(
            jacob_constr,
            constr,
            "jacobian_and_value",
            "jacob_constr",
            backend,
        )

    @cache_in_state("pos")
    def constr(self, state: ChainState) -> ArrayLike:
        return self._constr(state.pos)

    @cache_in_state_with_aux("pos", "constr")
    def jacob_constr(self, state: ChainState) -> ArrayLike:
        return self._jacob_constr(state.pos)

    def dh2_flow_dmom(
        self,
        state: ChainState,  # noqa: ARG002
        dt: ScalarLike,
    ) -> tuple[matrices.Matrix, matrices.Matrix]:
        return (dt * self.metric.inv, matrices.IdentityMatrix(self.metric.shape[0]))

    @cache_in_state("pos")
    def gram(self, state: ChainState) -> matrices.PositiveDefiniteMatrix:
        """Gram matrix at current position.

        The Gram matrix as a position `q` is defined as

        .. code::

            gram(q) = jacob_constr(q) @ inv(metric) @ jacob_constr(q).T

        where :code:`jacob_constr` is the Jacobian of the constraint function
        :code:`constr` and :code:`metric` is the matrix representation of the metric on
        the ambient space.

        Args:
            state: State to compute value at.

        Returns:
            Gram matrix as matrix object.
        """
        return self.jacob_constr_inner_product(
            self.jacob_constr(state),
            self.metric.inv,
        )

    def inv_gram(self, state: ChainState) -> matrices.PositiveDefiniteMatrix:
        """Inverse of Gram matrix at current position.

        Args:
            state: State to compute value at.

        Returns:
            Inverse of Gram matrix as matrix object.
        """
        return self.gram(state).inv

    def log_det_sqrt_gram(self, state: ChainState) -> ScalarLike:
        """Value of (half of) log-determinant of Gram matrix."""
        return 0.5 * self.gram(state).log_abs_det

    @abstractmethod
    def grad_log_det_sqrt_gram(self, state: ChainState) -> ArrayLike:
        """Derivative of half log-determinant of Gram matrix with respect to position.

        Args:
            state: State to compute value at.

        Returns:
            Value of :code:`log_det_sqrt_gram(state)` derivative with respect to
            :code:`state.pos`.
        """

    def h1(self, state: ChainState) -> ScalarLike:
        if self.dens_wrt_hausdorff:
            return self.neg_log_dens(state)
        return self.neg_log_dens(state) + self.log_det_sqrt_gram(state)

    def dh1_dpos(self, state: ChainState) -> ArrayLike:
        if self.dens_wrt_hausdorff:
            return self.grad_neg_log_dens(state)
        return self.grad_neg_log_dens(state) + self.grad_log_det_sqrt_gram(state)

    def project_onto_cotangent_space(
        self,
        mom: ArrayLike,
        state: ChainState,
    ) -> ArrayLike:
        # Use parenthesis to force right-to-left evaluation to avoid
        # matrix-matrix products
        mom -= self.jacob_constr(state).T @ (
            self.inv_gram(state) @ (self.jacob_constr(state) @ (self.metric.inv @ mom))
        )
        return mom


class DenseConstrainedEuclideanMetricSystem(ConstrainedEuclideanMetricSystem):
    r"""Euclidean Hamiltonian system subject to a dense set of constraints.

    See :py:class:`ConstrainedEuclideanMetricSystem` for more details about constrained
    systems.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        constr: ArrayFunction,
        *,
        metric: MetricLike | None = None,
        dens_wrt_hausdorff: bool = True,
        grad_neg_log_dens: GradientFunction | None = None,
        jacob_constr: JacobianFunction | None = None,
        mhp_constr: MatrixHessianProductFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the constrained
                position space with respect to the Hausdorff measure on the constraint
                manifold (if :code:`dens_wrt_hausdorff == True`) or alternatively the
                negative logarithm of an unnormalized probability density on the
                unconstrained (ambient) position space with respect to the Lebesgue
                measure. In the former case the target distribution it is wished to draw
                approximate samples from is assumed to be directly specified by the
                density function on the manifold. In the latter case the density
                function is instead taken to specify a prior distribution on the ambient
                space with the target distribution then corresponding to the posterior
                distribution when conditioning on the (zero Lebesgue measure) event
                :code:`constr(q) == 0` where :code:`q` is the position array. This
                target posterior distribution has support on the differentiable manifold
                implicitly defined by the constraint equation, with density with respect
                to the Hausdorff measure on the manifold corresponding to the ratio of
                the prior density (specified by :code:`neg_log_dens`) and the
                square-root of the determinant of the Gram matrix defined by
                :code:`gram(q) = jacob_constr(q) @ inv(metric) @ jacob_constr(q).T`
                where :code:`jacob_constr` is the Jacobian of the constraint function
                :code:`constr` and :code:`metric` is the matrix representation of the
                metric on the ambient space.
            constr: Function which given a position array return as a 1D array the value
                of the (vector-valued) constraint function, the zero level-set of which
                implicitly defines the manifold the dynamic is simulated on.
            metric: Matrix object corresponding to matrix representation of metric on
                *unconstrained* position space and covariance of Gaussian marginal
                distribution on *unconstrained* momentum vector. If :code:`None` is
                passed (the default), the identity matrix will be used. If a 1D array is
                passed then this is assumed to specify a metric with positive diagonal
                matrix representation and the array the matrix diagonal. If a 2D array
                is passed then this is assumed to specify a metric with a dense positive
                definite matrix representation specified by the array. Otherwise if the
                value is a :py:class:`mici.matrices.PositiveDefiniteMatrix` subclass it
                is assumed to directly specify the metric matrix representation.
            dens_wrt_hausdorff: Whether the :code:`neg_log_dens` function specifies the
                (negative logarithm) of the density of the target distribution with
                respect to the Hausdorff measure on the manifold directly (:code:`True`)
                or alternatively the negative logarithm of a density of a prior
                distriubtion on the unconstrained (ambient) position space with respect
                to the Lebesgue measure, with the target distribution then corresponding
                to the posterior distribution when conditioning on the event
                :code:`constr(pos) == 0` (:code:`False`). Note that in the former case
                the base Hausdorff measure on the manifold depends on the metric defined
                on the ambient space, with the Hausdorff measure being defined with
                respect to the metric induced on the manifold from this ambient metric.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of :code:`neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the :code:`neg_log_dens` evaluated at the
                passed position array. If :code:`None` is passed (the default) an
                automatic differentiation fallback will be used to attempt to construct
                a function to compute the derivative (and value) of :code:`neg_log_dens`
                automatically.
            jacob_constr: Function which given a position array computes the Jacobian
                (matrix / 2D array of partial derivatives) of the output of the
                constraint function :code:`c = constr(q)` with respect to the position
                array argument :code:`q`, returning the computed Jacobian as a 2D array
                :code:`jacob` with :code:`jacob[i, j] = ∂c[i] / ∂q[j]`. Optionally the
                function may instead return a 2-tuple of values with the first being the
                array corresponding to the Jacobian and the second being the value of
                :code:`constr` evaluated at the passed position array. If :code:`None`
                is passed (the default) an automatic differentiation fallback will be
                used to attempt to construct a function to compute the Jacobian (and
                value) of :code:`constr`
                automatically.
            mhp_constr: Function which given a position array returns another function
                which takes a 2D array as an argument and returns the
                *matrix-Hessian-product* (MHP) of the constraint function :code:`constr`
                with respect to the position array argument. The MHP is here defined as
                a function of a :code:`(dim_constr, dim_pos)` shaped 2D array :code:`m`
                as :code:`mhp(m) = sum(m[:, :, None] * hess[:, :, :], axis=(0, 1))`
                where :code:`hess` is the :code:`(dim_constr, dim_pos, dim_pos)` shaped
                vector-Hessian of :code:`c = constr(q)` with respect to :code:`q` i.e.
                the array of second-order partial derivatives of such that
                :code:`hess[i, j, k] = ∂²c[i] / (∂q[j] ∂q[k])`. Optionally the function
                may instead return a 3-tuple of values with the first a function to
                compute a MHP of :code:`constr`, the second a 2D array corresponding to
                the Jacobian of :code:`constr`, and the third the value of
                :code:`constr`, all evaluated at the passed position array. If
                :code:`None` is passed (the default) an automatic differentiation
                fallback will be used to attempt to construct a function which
                calculates the MHP (and Jacobian and value) of :code:`constr`
                automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            constr=constr,
            metric=metric,
            dens_wrt_hausdorff=dens_wrt_hausdorff,
            grad_neg_log_dens=grad_neg_log_dens,
            jacob_constr=jacob_constr,
            backend=backend,
        )
        if not dens_wrt_hausdorff:
            self._mhp_constr = autodiff_fallback(
                mhp_constr,
                constr,
                "mhp_jacobian_and_value",
                "mhp_constr",
                backend,
            )

    @cache_in_state_with_aux("pos", ("jacob_constr", "constr"))
    def mhp_constr(self, state: ChainState) -> MatrixHessianProduct:
        return self._mhp_constr(state.pos)

    def jacob_constr_inner_product(
        self,
        jacob_constr_1: MatrixLike,
        inner_product_matrix: matrices.PositiveDefiniteMatrix,
        jacob_constr_2: MatrixLike | None = None,
    ) -> MatrixLike:
        if jacob_constr_2 is None or jacob_constr_2 is jacob_constr_1:
            return matrices.DensePositiveDefiniteMatrix(
                jacob_constr_1 @ (inner_product_matrix @ jacob_constr_1.T),
            )
        return matrices.DenseSquareMatrix(
            jacob_constr_1 @ (inner_product_matrix @ jacob_constr_2.T),
        )

    @cache_in_state("pos")
    def grad_log_det_sqrt_gram(self, state: ChainState) -> ArrayLike:
        # Evaluate MHP of constraint function before Jacobian as Jacobian value
        # will potentially be computed in 'forward' pass and cached
        mhp_constr = self.mhp_constr(state)
        return mhp_constr(
            self.inv_gram(state) @ self.jacob_constr(state) @ self.metric.inv,
        )


class GaussianDenseConstrainedEuclideanMetricSystem(
    GaussianEuclideanMetricSystem,
    DenseConstrainedEuclideanMetricSystem,
):
    """Gaussian Euclidean Hamiltonian system subject to a dense set of constraints.

    See :py:class:`ConstrainedEuclideanMetricSystem` for more details about constrained
    systems and :py:class:`GaussianEuclideanMetricSystem` for Gaussian Euclidean metric
    systems.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        constr: ArrayFunction,
        *,
        metric: MetricLike | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        jacob_constr: JacobianFunction | None = None,
        mhp_constr: MatrixHessianProductFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the unconstrained
                (ambient) position space with respect to the standard Gaussian measure.
                The density function is taken to specify a prior distribution on the
                ambient space with the target distribution then corresponding to the
                posterior distribution when conditioning on the (zero Lebesgue measure)
                event `constr(pos) == 0`. This target posterior distribution has support
                on the differentiable manifold implicitly defined by the constraint
                equation, with density with respect to the Hausdorff measure on the
                manifold corresponding to the ratio of the prior density (specified by
                `neg_log_dens`) and the square-root of the determinant of the Gram
                matrix defined by.

                .. code-block::

                    gram(q) = jacob_constr(q) @ inv(metric) @ jacob_constr(q).T

                where `jacob_constr` is the Jacobian of the constraint function `constr`
                and `metric` is the matrix representation of the metric on the ambient
                space.
            constr: Function which given a position array return as a 1D array the value
                of the (vector-valued) constraint function, the zero level-set of which
                implicitly defines the manifold the dynamic is simulated on.
            metric: Matrix object corresponding to matrix representation of metric on
                *unconstrained* position space and covariance of Gaussian marginal
                distribution on *unconstrained* momentum vector. If `None` is passed
                (the default), the identity matrix will be used. If a 1D array is passed
                then this is assumed to specify a metric with positive diagonal matrix
                representation and the array the matrix diagonal. If a 2D array is
                passed then this is assumed to specify a metric with a dense positive
                definite matrix representation specified by the array. Otherwise if a
                subclass of `mici.matrices.PositiveDefiniteMatrix` it is assumed to
                directly specify the metric matrix representation.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct a function
                to compute the derivative (and value) of `neg_log_dens` automatically.
            jacob_constr: Function which given a position array computes the Jacobian
                (matrix / 2D array of partial derivatives) of the output of the
                constraint function `c = constr(q)` with respect to the position array
                argument `q`, returning the computed Jacobian as a 2D array `jacob` with

                .. code-block::

                    jacob[i, j] = ∂c[i] / ∂q[j]

                Optionally the function may instead return a 2-tuple of values with the
                first being the array corresponding to the Jacobian and the second being
                the value of `constr` evaluated at the passed position array. If `None`
                is passed (the default) an automatic differentiation fallback will be
                used to attempt to construct a function to compute the Jacobian (and
                value) of `neg_log_dens` automatically.
            mhp_constr: Function which given a position array returns another function
                which takes a 2D array as an argument and returns the
                *matrix-Hessian-product* (MHP) of the constraint function `constr` with
                respect to the position array argument. The MHP is here defined as a
                function of a `(dim_constr, dim_pos)` shaped 2D array `m`

                .. code-block::

                    mhp(m) = sum(m[:, :, None] * hess[:, :, :], axis=(0, 1))

                where `hess` is the `(dim_constr, dim_pos, dim_pos)` shaped
                vector-Hessian of `c = constr(q)` with respect to `q` i.e. the array of
                second-order partial derivatives of such that

                .. code-block::

                    hess[i, j, k] = ∂²c[i] / (∂q[j] ∂q[k])

                Optionally the function may instead return a 3-tuple of values with the
                first a function to compute a MHP of `constr`, the second a 2D array
                corresponding to the Jacobian of `constr`, and the third the value of
                `constr`, all evaluated at the passed position array. If `None` is
                passed (the default) an automatic differentiation fallback will be used
                to attempt to construct a function which calculates the MHP (and
                Jacobian and value) of `constr` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        DenseConstrainedEuclideanMetricSystem.__init__(
            self,
            neg_log_dens=neg_log_dens,
            constr=constr,
            metric=metric,
            dens_wrt_hausdorff=False,
            grad_neg_log_dens=grad_neg_log_dens,
            jacob_constr=jacob_constr,
            mhp_constr=mhp_constr,
            backend=backend,
        )

    def jacob_constr_inner_product(
        self,
        jacob_constr_1: MatrixLike,
        inner_product_matrix: matrices.PositiveDefiniteMatrix,
        jacob_constr_2: MatrixLike | None = None,
    ) -> MatrixLike:
        if jacob_constr_2 is None or jacob_constr_2 is jacob_constr_1:
            return matrices.DenseSymmetricMatrix(
                jacob_constr_1 @ (inner_product_matrix @ jacob_constr_1.T),
            )
        return matrices.DenseSquareMatrix(
            jacob_constr_1 @ (inner_product_matrix @ jacob_constr_2.T),
        )

    def dh2_flow_dmom(
        self,
        state: ChainState,  # noqa: ARG002
        dt: ScalarLike,
    ) -> tuple[matrices.Matrix, matrices.Matrix]:
        omega = 1.0 / self.metric.eigval**0.5
        sin_omega_dt, cos_omega_dt = np.sin(omega * dt), np.cos(omega * dt)
        return (
            matrices.EigendecomposedSymmetricMatrix(
                self.metric.eigvec,
                sin_omega_dt * omega,
            ),
            matrices.EigendecomposedSymmetricMatrix(self.metric.eigvec, cos_omega_dt),
        )


class RiemannianMetricSystem(System):
    r"""Riemannian Hamiltonian system with a position-dependent metric.

    This class allows for metric matrix representations of any generic type. In most
    cases a specialized subclass such as `DenseRiemannianMetricSystem`,
    `CholeskyFactoredRiemannianMetricSystem`, `DiagonalRiemannianMetricSystem`,
    `ScalarRiemannianMetricSystem` or `SoftAbsRiemannianMetricSystem` will provide a
    simpler method of constructng a system with a metric matrix representation of a
    specific type.

    The position space is assumed to be a Riemannian manifold with a metric with
    position-dependent positive definite matrix-representation :math:`M(q)` where
    :math:`q` is a position vector. The momentum :math:`p` is then taken to have a
    zero-mean Gaussian conditional distribution given the position :math:`q`, with
    covariance :math:`M(q)`, i.e. :math:`p \sim \mathcal{N}(0, M(q))` [1].

    The :math:`h_1` Hamiltonian component is then

    .. math::

        h_1(q) = \ell(q) + \frac{1}{2}\log\left|M(q)\right|

    where :math:`\ell(q)` is the negative log (unnormalized) density of the target
    distribution with respect to the Lebesgue measure at :math:`q`. The :math:`h_2`
    Hamiltonian component is

    .. math::

        h_2(q, p) = \frac{1}{2} p^T (M(q))^{-1} p.

    Due to the coupling between the position and momentum variables in :math:`h_2`, the
    Hamiltonian system is non-separable, requiring use of a numerical integrator with
    implicit steps when simulating the Hamiltonian dynamic associated with the system,
    e.g. `mici.integrators.ImplicitLeapfrogIntegrator`.

    References:
      1. Girolami, M. and Calderhead, B., 2011. Riemann manifold Langevin and
         Hamiltonian Monte Varlo methods. Journal of the Royal Statistical Society:
         Series B (Statistical Methodology), 73(2), pp.123-214.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        metric_matrix_class: type[matrices.PositiveDefiniteMatrix],
        metric_func: ArrayFunction,
        *,
        vjp_metric_func: VectorJacobianProductFunction | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        metric_kwargs: dict[str, Any] | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the Lebesgue measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            metric_matrix_class: Class (or factory function returning an instance of the
                class) which defines type of matrix representation of metric. The class
                initializer should take a single positional argument which will be
                passed the array outputted by `metric_func`, and which is assumed to be
                a parameter which fully defines the resulting matrix (e.g. the diagonal
                of a `mici.matrices.DiagonalMatrix`). The class initializer may also
                optionally take one or more keyword arguments, with the `metric_kwargs`
                argument used to specify the value of these, if any. Together this means
                the metric matrix representation at a position `pos` is constructed as.

                .. code-block::

                    metric = metric_matrix_class(metric_func(pos), **metric_kwargs)

                The `mici.matrices.PositiveDefiniteMatrix` subclass should as a minimum
                define `inv`, `log_abs_det`, `grad_log_abs_det`,
                `grad_quadratic_form_inv`, `__matmul__` and `__rmatmul__` methods /
                properties (see documentation of `mici.matrices.PositiveDefiniteMatrix`
                and `mici.matrices.DifferentiableMatrix` for definitions of the expected
                behaviour of these methods).
            metric_func: Function which given a position array returns an array
                containing the parameter value of the metric matrix representation
                passed as the single positional argument to the `metric_matrix_class`
                initializer.
            vjp_metric_func: Function which given a position array returns another
                function which takes an array as an argument and returns the
                **vector-Jacobian-product** (VJP) of `metric_func` with respect to the
                position array argument. The VJP is here defined as a function of an
                array `v` (of the same shape as the output of `metric_func`)
                corresponding to

                .. code-block::

                    vjp(v) = sum(v[..., None] * jacob, tuple(range(v.ndim))

                where `jacob` is the Jacobian of `m = metric_func(q)` wrt `q` i.e. the
                array of partial derivatives of the function such that

                .. code-block::

                    jacob[..., i] = ∂m[...] / ∂q[i]

                Optionally the function may instead return a 2-tuple of values with the
                first a function to compute a VJP of `metric_func` and the second an
                array containing the value of `metric_func`, both evaluated at the
                passed position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct a function
                which calculates the VJP (and value) of `metric_func` automatically.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            metric_kwargs: An optional dictionary of any additional keyword arguments to
                the initializer of `metric_matrix_class`.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        self._metric_matrix_class = metric_matrix_class
        self._metric_func = wrap_function(metric_func, backend)
        self._vjp_metric_func = autodiff_fallback(
            vjp_metric_func,
            metric_func,
            "vjp_and_value",
            "vjp_metric_func",
            backend,
        )
        self._metric_kwargs = {} if metric_kwargs is None else metric_kwargs
        super().__init__(neg_log_dens, grad_neg_log_dens=grad_neg_log_dens)

    @cache_in_state("pos")
    def metric_func(self, state: ChainState) -> ArrayLike:
        """Function computing the parameter of the metric matrix representation.

        Args:
            state: State to compute value at.

        Returns:
            Value of `metric_func(state.pos)`.
        """
        return self._metric_func(state.pos)

    @cache_in_state_with_aux("pos", "metric_func")
    def vjp_metric_func(self, state: ChainState) -> VectorJacobianProduct:
        """Function constructing a vector-Jacobian-product for `metric_func`.

        The vector-Jacobian-product is here defined as a function of an array `v` (of
        the same shape as the output of `metric_func`) corresponding to

        .. code-block::

            vjp(v) = sum(v[..., None] * jacob, axis=tuple(range(v.ndim))

        where `jacob` is the Jacobian of `m = metric_func(q)` wrt `q` i.e. the array of
        partial derivatives of the function such that

        .. code-block::

            jacob[..., i] = ∂m[...] / ∂q[i]

        Args:
            state: State to compute VJP at.

        Returns:
            Vector-Jacobian-product function.
        """
        return self._vjp_metric_func(state.pos)

    @cache_in_state("pos")
    def metric(self, state: ChainState) -> matrices.PositiveDefiniteMatrix:
        """Function computing the metric matrix representation.

        The returned type of this function is that specified by the
        `metric_matrix_class` argument to the initializer.

        Args:
            state: State to compute value at.

        Returns:
            Metric matrix representation.
        """
        return self._metric_matrix_class(self.metric_func(state), **self._metric_kwargs)

    def h(self, state: ChainState) -> ScalarLike:
        return self.h1(state) + self.h2(state)

    def h1(self, state: ChainState) -> ScalarLike:
        return self.neg_log_dens(state) + 0.5 * self.metric(state).log_abs_det

    def dh1_dpos(self, state: ChainState) -> ArrayLike:
        # Evaluate VJP of metric function before metric as metric value will
        # potentially be computed in forward pass and cached
        vjp_metric = self.vjp_metric_func(state)
        return self.grad_neg_log_dens(state) + 0.5 * vjp_metric(
            self.metric(state).grad_log_abs_det,
        )

    def h2(self, state: ChainState) -> ScalarLike:
        return 0.5 * state.mom @ self.metric(state).inv @ state.mom

    def dh2_dpos(self, state: ChainState) -> ArrayLike:
        # Evaluate VJP of metric function before metric as metric value will
        # potentially be computed in forward pass and cached
        vjp_metric = self.vjp_metric_func(state)
        return 0.5 * vjp_metric(self.metric(state).grad_quadratic_form_inv(state.mom))

    def dh2_dmom(self, state: ChainState) -> ArrayLike:
        return self.metric(state).inv @ state.mom

    def sample_momentum(self, state: ChainState, rng: Generator) -> ArrayLike:
        return self.metric(state).sqrt @ rng.normal(size=state.pos.shape)


class ScalarRiemannianMetricSystem(RiemannianMetricSystem):
    """Riemannian-metric system with scaled identity matrix representation.

    Hamiltonian system with a position dependent scaled identity metric matrix
    representation which is specified by a scalar function `metric_scalar_function` of
    the position `q` which outputs a strictly positive scalar `s =
    metric_scalar_func(q)` with the the metric matrix representation then taken to be `s
    * identity(q.shape[0])`.

    See documentation of `RiemannianMetricSystem` for more general details about
    Riemannian-metric Hamiltonian systems.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        metric_scalar_func: ScalarFunction,
        *,
        vjp_metric_scalar_func: VectorJacobianProductFunction | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the Lebesgue measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            metric_scalar_func: Function which given a position array returns a strictly
                positive scalar corresponding to the parameter value of the scaled
                identity metric matrix representation.
            vjp_metric_scalar_func: Function which given a position array returns
                another function which takes a scalar as an argument and returns the
                *vector-Jacobian-product* (VJP) of `metric_scalar_func` with respect to
                the position array argument. The VJP is here defined as a function of a
                scalar `v`.

                .. code-block::

                    vjp(v) = v @ grad

                where `grad` is the `(dim_pos,)` shaped Jacobian (gradient) of `s =
                metric_scalar_func(q)` with respect to `q` i.e. the array of partial
                derivatives of the function such that

                .. code-block::

                    grad[i] = ∂s / ∂q[i]

                Optionally the function may instead return a 2-tuple of values with the
                first a function to compute a VJP of `metric_scalar_func` and the second
                a float containing the value of `metric_scalar_func`, both evaluated at
                the passed position array. If `None` is passed (the default) an
                automatic differentiation fallback will be used to attempt to construct
                a function which calculates the VJP (and value) of `metric_scalar_func`
                automatically.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            metric_matrix_class=matrices.PositiveScaledIdentityMatrix,
            metric_func=metric_scalar_func,
            vjp_metric_func=vjp_metric_scalar_func,
            grad_neg_log_dens=grad_neg_log_dens,
            backend=backend,
        )

    @cache_in_state("pos")
    def metric(self, state: ChainState) -> matrices.PositiveDefiniteMatrix:
        return self._metric_matrix_class(
            self.metric_func(state),
            size=state.pos.shape[0],
        )


class DiagonalRiemannianMetricSystem(RiemannianMetricSystem):
    """Riemannian-metric system with diagonal matrix representation.

    Hamiltonian system with a position dependent diagonal metric matrix representation
    which is specified by a vector-valued function `metric_diagonal_func` of the
    position `q` which outputs a 1D array with strictly positive elements `d =
    metric_diagonal_func(q)` with the metric matrix representation then taken to be
    `diag(d)`.

    See documentation of `RiemannianMetricSystem` for more general details about
    Riemannian-metric Hamiltonian systems.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        metric_diagonal_func: ArrayFunction,
        *,
        vjp_metric_diagonal_func: VectorJacobianProductFunction | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the Lebesgue measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            metric_diagonal_func: Function which given a position array returns a 1D
                array with strictly positive values corresponding to the diagonal values
                (left-to-right) of the diagonal metric matrix representation.
            vjp_metric_diagonal_func: Function which given a position array returns
                another function which takes a 1D array as an argument and returns the
                *vector-Jacobian-product* (VJP) of `metric_diagonal_func` with respect
                to the position array argument. The VJP is here defined as a function of
                a 1D array `v`.

                .. code-block::

                    vjp(v) = sum(v[:, None] * jacob[:, :], axis=0)

                where `jacob` is the `(dim_pos, dim_pos)` shaped Jacobian of `d =
                metric_diagonal_func(q)` with respect to `q` i.e. the array of partial
                derivatives of the function such that

                .. code-block::

                    jacob[i, j] = ∂d[i] / ∂q[j]

                Optionally the function may instead return a 2-tuple of values with the
                first a function to compute a VJP of `metric_diagonal_func` and the
                second a 1D array containing the value of `metric_diagonal_func`, both
                evaluated at the passed position array. If `None` is passed (the
                default) an automatic differentiation fallback will be used to attempt
                to construct a function which calculates the VJP (and value) of
                `metric_diagonal_func` automatically.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            metric_matrix_class=matrices.PositiveDiagonalMatrix,
            metric_func=metric_diagonal_func,
            vjp_metric_func=vjp_metric_diagonal_func,
            grad_neg_log_dens=grad_neg_log_dens,
            backend=backend,
        )


class CholeskyFactoredRiemannianMetricSystem(RiemannianMetricSystem):
    """Riemannian-metric system with Cholesky-factored matrix representation.

    Hamiltonian system with a position dependent metric matrix representation which is
    specified by its Cholesky factor by a matrix function `metric_chol_func` of the
    position `q` which outputs a lower-triangular matrix `L = metric_chol_func(q)` with
    the metric matrix representation then taken to be `L @ L.T`.

    See documentation of `RiemannianMetricSystem` for more general details about
    Riemannian-metric Hamiltonian systems.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        metric_chol_func: ArrayFunction,
        *,
        vjp_metric_chol_func: VectorJacobianProductFunction | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the Lebesgue measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            metric_chol_func: Function which given a position array returns a 2D array
                with zeros above the diagonal corresponding to the lower-triangular
                Cholesky-factor of the positive definite metric matrix representation.
            vjp_metric_chol_func: Function which given a position array returns another
                function which takes a lower-triangular 2D array as an argument (any
                values in the array above the diagonal are ignored) and returns the
                *vector-Jacobian-product* (VJP) of `metric_chol_func` with respect to
                the position array argument. The VJP is here defined as a function of a
                2D array `v`.

                .. code-block::

                    vjp(v) = sum(v[:, :, None] * jacob[:, :, :], axis=(0, 1))

                where `jacob` is the `(dim_pos, dim_pos, dim_pos)` shaped Jacobian of `L
                = metric_chol_func(q)` with respect to `q` i.e. the array of partial
                derivatives of the function such that

                .. code-block::

                    jacob[i, j, k] = ∂L[i, j] / ∂q[k]

                Optionally the function may instead return a 2-tuple of values with the
                first a function to compute a VJP of `metric_chol_func` and the second a
                2D array containing the value of `metric_chol_func`, both evaluated at
                the passed position array. If `None` is passed (the default) an
                automatic differentiation fallback will be used to attempt to construct
                a function which calculates the VJP (and value) of `metric_chol_func`
                automatically.
            grad_neg_log_dens:  Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            metric_matrix_class=matrices.TriangularFactoredPositiveDefiniteMatrix,
            metric_func=metric_chol_func,
            vjp_metric_func=vjp_metric_chol_func,
            grad_neg_log_dens=grad_neg_log_dens,
            metric_kwargs={"factor_is_lower": True},
            backend=backend,
        )


class DenseRiemannianMetricSystem(RiemannianMetricSystem):
    """Riemannian-metric system with dense matrix representation.

    Hamiltonian system with a position dependent metric matrix representation which is
    specified to be a dense matrix function `metric_func` of the position `q` which is
    guaranteed to be positive definite almost-everywhere, with `M = metric_func(q)` then
    the metric matrix representation.

    See documentation of `RiemannianMetricSystem` for more general details about
    Riemannian-metric Hamiltonian systems.
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        metric_func: ArrayFunction,
        *,
        vjp_metric_func: VectorJacobianProductFunction | None = None,
        grad_neg_log_dens: GradientFunction | None = None,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the Lebesgue measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            metric_func: Function which given a position array returns a 2D array
                corresponding to the positive definite metric matrix representation. The
                returned matrices (2D arrays) are assumed to be positive-definite for
                all input positions and a `LinAlgError` exception may be raised if this
                fails to be the case.
            vjp_metric_func: Function which given a position array returns another
                function which takes a 2D array as an argument and returns the
                *vector-Jacobian-product* (VJP) of `metric_func` with respect to the
                position array argument. The VJP is here defined as a function of a 2D
                array `v`.

                .. code-block::

                    vjp(v) = sum(v[:, :, None] * jacob[:, :, :], axis=(0, 1))

                where `jacob` is the `(dim_pos, dim_pos, dim_pos)` shaped Jacobian of `M
                = metric_func(q)` with respect to `q` i.e. the array of partial
                derivatives of the function such that

                .. code-block::

                    jacob[i, j, k] = ∂M[i, j] / ∂q[k]

                Optionally the function may instead return a 2-tuple of values with the
                first a function to compute a VJP of `metric_func` and the second a 2D
                array containing the value of `metric_func`, both evaluated at the
                passed position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct a function
                which calculates the VJP (and value) of `metric_func` automatically.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        super().__init__(
            neg_log_dens=neg_log_dens,
            metric_matrix_class=matrices.DensePositiveDefiniteMatrix,
            metric_func=metric_func,
            vjp_metric_func=vjp_metric_func,
            grad_neg_log_dens=grad_neg_log_dens,
            backend=backend,
        )


class SoftAbsRiemannianMetricSystem(RiemannianMetricSystem):
    """SoftAbs Riemmanian metric Hamiltonian system.

    Hamiltonian system with a position dependent metric matrix representation which is
    specified to be a dense matrix function `metric_func` of the position `q` which is
    guaranteed to be positive definite almost-everywhere, with `M = metric_func(q)` then
    the metric matrix representation.

    Hamiltonian system with a position dependent metric matrix representation which is
    specified to be an eigenvalue-regularized transformation of the Hessian of the
    negative log density function (the symmetric matrix of second derivatives the
    negative log density function with respect to the position array components.
    Specifically if `hess_neg_log_dens` is a symmetric 2D square array valued function
    of the position `q`, with `H = hess_neg_log_dens(q)` then if an eigenvalue
    decomposition of `H` is computed, i.e. `eigval, eigvec = eigh(H)`, with `eigval` a
    1D array of real eigenvalues, and `eigvec` the corresponding 2D array (orthogonal
    matrix) with eigenvectors as columns, then the resulting positive-definite metric
    matrix representation `M` is computed as

    .. code-block::

        M = eigvec @ diag(softabs(eigval, softabs_coeff)) @ eigvec.T

    with `softabs(x, softabs_coeff) = x / tanh(x * softabs_coeff)` an elementwise
    function which acts as a smooth approximation to the absolute function (ensuring all
    the eigenvalues of `M` are strictly positive) with the additional scalar parameter
    `softabs_coeff` controlling the smoothness of the approximation, with `softabs`
    tending to the piecewise linear `abs` function as `softabs_coeff` tends to infinity,
    and becoming increasingly smooth as `softabs_coeff` tends to zero.

    See documentation of `RiemannianMetricSystem` for more general details about
    Riemannian-metric Hamiltonian systems.

    References:
      1. Betancourt, M., 2013. A general metric for Riemannian manifold Hamiltonian
         Monte Carlo. In Geometric science of information (pp. 327-334).
    """

    def __init__(
        self,
        neg_log_dens: ScalarFunction,
        *,
        grad_neg_log_dens: GradientFunction | None = None,
        hess_neg_log_dens: HessianFunction | None = None,
        mtp_neg_log_dens: MatrixTressianProductFunction | None = None,
        softabs_coeff: ScalarLike = 1.0,
        backend: str | None = None,
    ) -> None:
        """
        Args:
            neg_log_dens: Function which given a position array returns the negative
                logarithm of an unnormalized probability density on the position space
                with respect to the Lebesgue measure, with the corresponding
                distribution on the position space being the target distribution it is
                wished to draw approximate samples from.
            grad_neg_log_dens: Function which given a position array returns the
                derivative of `neg_log_dens` with respect to the position array
                argument. Optionally the function may instead return a 2-tuple of values
                with the first being the array corresponding to the derivative and the
                second being the value of the `neg_log_dens` evaluated at the passed
                position array. If `None` is passed (the default) an automatic
                differentiation fallback will be used to attempt to construct the
                derivative of `neg_log_dens` automatically.
            hess_neg_log_dens: Function which given a position array returns the Hessian
                of `neg_log_dens` with respect to the position array argument as a 2D
                array. Optionally the function may instead return a 3-tuple of values
                with the first a 2D array containting the Hessian of `neg_log_dens`, the
                second a 1D array containing the gradient of `neg_log_dens` and the
                third the value of `neg_log_dens`, all evaluated at the passed position
                array. If `None` is passed (the default) an automatic differentiation
                fallback will be used to attempt to construct a function which
                calculates the Hessian (and gradient and value) of `neg_log_dens`
                automatically.
            mtp_neg_log_dens: Function which given a position array returns another
                function which takes a 2D array (matrix) as an argument and returns the
                *matrix-Tressian-product* (MTP) of `neg_log_dens` with respect to the
                position array argument. The MTP is here defined as a function of a
                matrix `m` corresponding to.

                .. code-block::

                    mtp(m) = sum(m[:, :, None] * tress[:, :, :], axis=(0, 1))

                where `tress` is the 'Tressian' of `f = neg_log_dens(q)` wrt `q` i.e.
                the 3D array of third-order partial derivatives of the scalar-valued
                function such that

                .. code-block::

                    tress[i, j, k] = ∂³f / (∂q[i] ∂q[j] ∂q[k])

                Optionally the function may instead return a 4-tuple of values with the
                first a function to compute a MTP of `neg_log_dens`, the second a 2D
                array containing the Hessian of `neg_log_dens`, the third a 1D array
                containing the gradient of `neg_log_dens` and the fourth the value of
                `neg_log_dens`, all evaluated at the passed position array. If `None` is
                passed (the default) an automatic differentiation fallback will be used
                to attempt to construct a function which calculates the MTP (and Hesisan
                and gradient and value) of `neg_log_dens` automatically.
            softabs_coeff: Positive regularisation coefficient for smooth approximation
                to absolute value used to regularize Hessian eigenvalues in metric
                matrix representation. As the value tends to infinity the approximation
                becomes increasingly close to the absolute function.
            backend: Name of automatic differentiation backend to use. See
                :py:mod:`.autodiff` subpackage documentation for details of available
                backends. If `None` (the default) no automatic differentiation fallback
                will be used and so all derivative functions must be specified
                explicitly.
        """
        self._hess_neg_log_dens = autodiff_fallback(
            hess_neg_log_dens,
            neg_log_dens,
            "hessian_grad_and_value",
            "neg_log_dens",
            backend,
        )
        self._mtp_neg_log_dens = autodiff_fallback(
            mtp_neg_log_dens,
            neg_log_dens,
            "mtp_hessian_grad_and_value",
            "mtp_neg_log_dens",
            backend,
        )
        super().__init__(
            neg_log_dens=neg_log_dens,
            metric_matrix_class=matrices.SoftAbsRegularizedPositiveDefiniteMatrix,
            metric_func=self._hess_neg_log_dens,
            vjp_metric_func=self._mtp_neg_log_dens,
            grad_neg_log_dens=grad_neg_log_dens,
            metric_kwargs={"softabs_coeff": softabs_coeff},
            backend=backend,
        )

    def metric_func(self, state: ChainState) -> ArrayLike:
        return self.hess_neg_log_dens(state)

    def vjp_metric_func(self, state: ChainState) -> MatrixTressianProduct:
        return self.mtp_neg_log_dens(state)

    @cache_in_state_with_aux("pos", ("grad_neg_log_dens", "neg_log_dens"))
    def hess_neg_log_dens(self, state: ChainState) -> ArrayLike:
        """Hessian of negative log density with respect to position.

        Args:
            state: State to compute value at.

        Returns:
            2D array of `neg_log_dens(state)` second derivatives with respect to
            `state.pos`, with `hessian[i, j]` the second derivative of
            `neg_log_dens(state)` with respect to `state.pos[i]` and `state.pos[j]`.
        """
        return self._hess_neg_log_dens(state.pos)

    @cache_in_state_with_aux(
        "pos",
        ("hess_neg_log_dens", "grad_neg_log_dens", "neg_log_dens"),
    )
    def mtp_neg_log_dens(self, state: ChainState) -> MatrixTressianProduct:
        """Generate MTP of negative log density with respect to position.

        The matrix-Tressian-product (MTP) is here defined as a function of a matrix `m`
        corresponding to

        .. code-block::

            mtp(m) = sum(m[:, :, None] * tress[:, :, :], axis=(0, 1))

        where `tress` is the 'Tressian' of `f = neg_log_dens(q)` with respect to `q =
        state.pos` i.e. the 3D array of third-order partial derivatives of the
        scalar-valued function such that

        .. code-block::

            tress[i, j, k] = ∂³f / (∂q[i] ∂q[j] ∂q[k])

        Args:
            state: State to compute value at.

        Returns:
            Function which accepts a 2D array of shape
            `(state.pos.shape[0], state.pos.shape[0])` as an argument and returns an
            array of shape `state.pos.shape` containing the computed MTP value.
        """
        return self._mtp_neg_log_dens(state.pos)



"""Markov transition kernels."""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod, abstractproperty
from typing import TYPE_CHECKING, NamedTuple

import numpy as np

from mici.errors import (
    ConvergenceError,
    Error,
    HamiltonianDivergenceError,
    IntegratorError,
    NonReversibleStepError,
)
from mici.utils import LogRepFloat

if TYPE_CHECKING:
    from numpy.random import Generator
    from numpy.typing import ArrayLike, DTypeLike

    from mici.integrators import Integrator
    from mici.states import ChainState
    from mici.systems import System
    from mici.types import ScalarLike, TerminationCriterion


logger = logging.getLogger(__name__)


def _process_integrator_error(exception: Error, stats: dict[str, ScalarLike]) -> None:
    logger.info(f"Terminating trajectory due to error:\n{exception!s}")
    # Only set stats fields to True if exception is of matching type.
    # Corresponding fields should be set to False by default for transitions
    # which potentially raise these errors.
    if isinstance(exception, HamiltonianDivergenceError):
        stats["diverging"] = True
    elif isinstance(exception, NonReversibleStepError):
        stats["non_reversible_step"] = True
    elif isinstance(exception, ConvergenceError):
        stats["convergence_error"] = True


class Transition(ABC):
    """Base class for Markov transition kernels.

    Defines expected interface for transitions by sampler classes.
    """

    @abstractproperty
    def state_variables(self) -> set[str]:
        """A set of names of state variables accessed by this transition."""

    @property
    def statistic_types(self) -> dict[str, tuple[DTypeLike, ScalarLike]] | None:
        """A dictionary describing the statistics computed during transition.

        Either :code:`None` if no statistics are returned by :py:meth:`sample` method or
        a dictionary with string keys and tuple values, with the keys defining the keys
        of the statistics returned in the :code:`trans_stats` return value of the
        :py:meth:`sample` method and the first entry of the value tuples an appropriate
        NumPy `dtype` for the array used to store the corresponding statistic values and
        second entry the default value to initialize this array with.
        """
        return None

    @abstractmethod
    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike] | None]:
        """Sample a new chain state from the Markov transition kernel.

        Args:
            state: Current chain state to condition transition kernel on.
            rng: Numpy random number generator.

        Returns:
            Tuple of updated state object and any statistics computed during the
            transition or :code:`None` if no statistics.
        """


class MomentumTransition(Transition):
    """Base class for momentum transitions.

    Markov transition  kernel which leaves the conditional distribution on the momentum
    under the canonical distribution invariant, updating only the momentum component of
    the chain state.
    """

    @property
    def state_variables(self) -> set[str]:
        return {"mom"}

    def __init__(self, system: System) -> None:
        """
        Args:
            system: Hamiltonian system defining conditional distribution on momentum to
                leave invariant.
        """
        self.system = system

    @abstractmethod
    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike] | None]:
        """Sample a new momentum component to state.

        Assigns a new momentum component to state by sampling from a Markov transition
        kernel which leaves the conditional distribution on the momentum under the
        canonical distribution defined by the Hamiltonian system invariant.

        Args:
            state: Current chain state to condition transition kernel on.
            rng: Numpy random number generator.

        Returns:
            Tuple of updated state object and any statistics computed during the
            transition or `None` if no statistics.
        """


class IndependentMomentumTransition(MomentumTransition):
    """Independent momentum transition.

    Independently resamples the momentum component of the state from its conditional
    distribution given the remaining state.
    """

    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike] | None]:
        state.mom = self.system.sample_momentum(state, rng)
        return state, None


class CorrelatedMomentumTransition(MomentumTransition):
    """Correlated (partial) momentum transition.

    Rather than independently sampling a new momentum, instead a pertubative
    Crank-Nicolson type update which produces a new momentum value with a specified
    correlation with the previous value is used. It is assumed that the conditional
    distribution of the momenta is zero-mean Gaussian such that the Crank-Nicolson
    update leaves the momenta conditional distribution exactly invariant. This approach
    is sometimes known as partial momentum refreshing or updating, and was originally
    proposed in (Horowitz, 1991).

    If the resampling coefficient is equal to zero then the momentum is not randomized
    at all and succesive applications of the coupled integration transitions will
    continue along the same simulated Hamiltonian trajectory. When an integration
    transition is accepted this means the subsequent simulated trajectory will continue
    evolving in the same direction and so not randomising the momentum will reduce
    random-walk behaviour. However on a rejection the integration direction is reversed
    and so without randomisation the trajectory will exactly backtrack along the
    previous tractory states. A resampling coefficient of one corresponds to the
    standard case of independent resampling of the momenta while intermediate values
    between zero and one correspond to varying levels of correlation between the pre and
    post update momentums.

    References:
      1. Horowitz, A.M. (1991). A generalized guided Monte Carlo algorithm. Phys. Lett.
         B, 268(CERN-TH-6172-91), pp.247-252.
    """

    def __init__(self, system: System, mom_resample_coeff: float = 1.0) -> None:
        """
        Args:
            system: Hamiltonian system defining conditional distribution on momentum to
                leave invariant.
            mom_resample_coeff: Scalar value in [0, 1] defining the momentum resampling
                coefficient.
        """
        super().__init__(system)
        if not (mom_resample_coeff >= 0 and mom_resample_coeff <= 1):
            msg = "mom_resample_coeff should have a value in the interval [0, 1]."
            raise ValueError(msg)
        self.mom_resample_coeff = mom_resample_coeff

    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike] | None]:
        if state.mom is None or self.mom_resample_coeff == 1:
            state.mom = self.system.sample_momentum(state, rng)
        elif self.mom_resample_coeff != 0:
            mom_ind = self.system.sample_momentum(state, rng)
            state.mom *= (1.0 - self.mom_resample_coeff**2) ** 0.5
            state.mom += self.mom_resample_coeff * mom_ind
        return state, None


class IntegrationTransition(Transition):
    """Base class for integration transtions.

    Markov transition kernel which leaves canonical distribution invariant and jointly
    updates the position and momentum components of the chain state by integrating the
    Hamiltonian dynamics of the system to propose new values for the state.
    """

    @property
    def state_variables(self) -> set[str]:
        return {"pos", "mom", "dir"}

    @property
    def statistic_types(self) -> dict[str, tuple[DTypeLike, ScalarLike]]:
        return self._statistic_types

    def __init__(self, system: System, integrator: Integrator) -> None:
        """
        Args:
            system: Hamiltonian system to be simulated.
            integrator: Symplectic integrator appropriate to the specified Hamiltonian
                system.
        """
        self.system = system
        self.integrator = integrator
        self._statistic_types = {
            "n_step": (np.int64, -1),
            "accept_stat": (np.float64, np.nan),
            "non_reversible_step": (bool, False),
            "convergence_error": (bool, False),
            "step_size": (np.float64, np.nan),
        }

    @abstractmethod
    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike] | None]:
        """Sample a position-momentum pair using integration based proposal(s).

        Samples new position and momentum values from a Markov transition kernel which
        leaves the canonical distribution on the state space corresponding to the
        Hamiltonian system invariant.

        Args:
            state: Current chain state.
            rng: Numpy random number generator.

        Returns:
            Tuple of updated state object and any statistics computed during the
            transition or `None` if no statistics.
        """


class MetropolisIntegrationTransition(IntegrationTransition):
    """Base for HMC methods using a Metropolis accept step to sample new state.

    In each transition a trajectory is generated by integrating the Hamiltonian dynamics
    from the current state in the current integration time direction for a number of
    integrator steps.

    The state at the end of the trajectory with the integration direction negated (this
    ensuring the proposed move is an involution) is used as the proposal in a Metropolis
    acceptance step. The integration direction is then deterministically negated again
    irrespective of the accept decision, with the effect being that on acceptance the
    integration direction will be equal to its initial value and on rejection the
    integration direction will be the negation of its initial value.
    """

    def __init__(self, system: System, integrator: Integrator) -> None:
        super().__init__(system, integrator)
        self._statistic_types["metrop_accept_prob"] = (np.float64, np.nan)

    def _sample_n_step(
        self,
        state: ChainState,
        n_step: int,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike]]:
        h_init = self.system.h(state)
        state_p = state
        integration_error = False
        stats = {
            "convergence_error": False,
            "non_reversible_step": False,
            "step_size": self.integrator.step_size,
        }
        try:
            for _s in range(n_step):
                state_p = self.integrator.step(state_p)
        except IntegratorError as e:
            integration_error = True
            stats["n_step"] = _s
            _process_integrator_error(e, stats)
        else:
            stats["n_step"] = n_step
            # Reverse integration direction of proposal to form an involution
            state_p.dir *= -1
        if state_p is not state:
            h_final = self.system.h(state_p)
            h_diff = h_init - h_final
            # Explicitly check if h_diff is NaN as min(0, NaN) = 0
            accept_prob = 0.0 if np.isnan(h_diff) else np.exp(min(0, h_diff))
        else:
            accept_prob = 0.0
        stats["metrop_accept_prob"] = accept_prob
        stats["accept_stat"] = accept_prob if not integration_error else 0.0
        if not integration_error and rng.uniform() < accept_prob:
            state = state_p
        # Reverse integration direction of new state
        # As extended target distribution is symmetric in direction indicator
        # this always leaves the distribution invariant
        state.dir *= -1
        return state, stats


class MetropolisStaticIntegrationTransition(MetropolisIntegrationTransition):
    """Static integration transition with Metropolis sampling of new state.

    In this variant the trajectory is generated by integrating the state through time a
    fixed number of integrator steps. This is original proposed Hybrid Monte Carlo
    (often now instead termed Hamiltonian Monte Carlo) algorithm (Duane et al., 1987;
    Neal, 2011).

    References:
      1. Duane, S., Kennedy, A.D., Pendleton, B.J. and Roweth, D. (1987). Hybrid Monte
         Carlo. Physics letters B, 195(2), pp.216-222.
      2. Neal, R.M. (2011). MCMC using Hamiltonian dynamics. Handbook of Markov Chain
         Monte Carlo, 2(11), p.2.
    """

    def __init__(self, system: System, integrator: Integrator, n_step: int) -> None:
        """
        Args:
            system: Hamiltonian system to be simulated.
            integrator: Symplectic integrator appropriate to the specified Hamiltonian
                system.
            n_step: Number of integrator steps to simulate in each transition.
        """
        super().__init__(system, integrator)
        if n_step <= 0:
            msg = "Number of integrator steps must be positive."
            raise ValueError(msg)
        self.n_step = n_step

    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike] | None]:
        return self._sample_n_step(state, self.n_step, rng)


class MetropolisRandomIntegrationTransition(MetropolisIntegrationTransition):
    """Random integration transition with Metropolis sampling of new state.

    In each transition a trajectory is generated by integrating the state in the current
    integration direction in time a random integer number of integrator steps sampled
    from the uniform distribution on an integer interval. The randomisation of the
    number of integration steps avoids the potential of the chain mixing poorly due to
    using an integration time close to the period of (near) periodic systems
    (Neal, 2011; Mackenzie, 1989).

    References:
      1. Neal, R.M. (2011). MCMC using Hamiltonian dynamics. Handbook of Markov Chain
         Monte Carlo, 2(11), p.2.
      2. Mackenzie, P.B. (1989). An improved hybrid Monte Carlo method. Physics Letters
         B, 226(3-4), pp.369-371.
    """

    def __init__(
        self,
        system: System,
        integrator: Integrator,
        n_step_range: tuple[int, int],
    ) -> None:
        """
        Args:
            system: Hamiltonian system to be simulated.
            integrator: Symplectic integrator appropriate to the specified Hamiltonian
                system.
            n_step_range: Tuple :code:`(lower, upper)` with two positive integer entries
                :code:`lower` and :code:`upper` (with :code:`upper > lower`) specifying
                respectively the lower and upper bounds (inclusive) of integer interval
                to uniformly draw random number integrator steps to simulate in each
                transition.
        """
        super().__init__(system, integrator)
        n_step_lower, n_step_upper = n_step_range
        if not (n_step_lower > 0 and n_step_lower < n_step_upper):
            msg = "Range bounds must be non-negative and first entry less than last."
            raise ValueError(msg)
        self.n_step_range = n_step_range

    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike] | None]:
        n_step = rng.integers(*self.n_step_range)
        return self._sample_n_step(state, n_step, rng)


def euclidean_no_u_turn_criterion(
    system: System,
    state_1: ChainState,
    state_2: ChainState,
    _sum_mom: ArrayLike,
) -> bool:
    """No-U-turn termination criterion for Euclidean manifolds.

    Terminates trajectories when the velocities at the terminal states of the trajectory
    both have negative dot products with the vector from the position of the first
    terminal state to the position of the second terminal state, corresponding to
    further evolution of the trajectory reducing the distance between the terminal state
    positions.

    Args:
        system: Hamiltonian system being integrated.
        state_1: First terminal state of trajectory.
        state_2: Second terminal state of trajectory.
        _sum_mom: Sum of momentums of trajectory states (unused).

    Returns:
        Whether termination criterion is satisfied.

    References:
      1. Hoffman, M.D. and Gelman, A. (2014). The No-U-turn sampler: adaptively setting
         path lengths in Hamiltonian Monte Carlo. Journal of Machine Learning Research,
         15(1), pp.1593-1623.
    """
    return (
        np.sum(system.dh_dmom(state_1) * (state_2.pos - state_1.pos)) < 0
        or np.sum(system.dh_dmom(state_2) * (state_2.pos - state_1.pos)) < 0
    )


def riemannian_no_u_turn_criterion(
    system: System,
    state_1: ChainState,
    state_2: ChainState,
    sum_mom: ArrayLike,
) -> bool:
    """Generalized no-U-turn termination criterion on Riemannian manifolds.

    Terminates trajectories when the velocities at the terminal states of the trajectory
    both have negative dot products with the sum of the the momentums across the
    trajectory from the first to second terminal state of the first terminal state to
    the position of the second terminal state (Betancourt, 2013). This generalizes the
    no-U-turn criterion of Hoffman and Gelman (2014) to Riemannian manifolds where due
    to the intrinsic curvature of the space the geodesic between two points is general
    no longer a straight line.

    Args:
        system: Hamiltonian system being integrated.
        state_1: First terminal state of trajectory.
        state_2: Second terminal state of trajectory.
        sum_mom: Sum of momentums of trajectory states.

    Returns:
        Whether termination criterion is satisfied.

    References:
      1. Hoffman, M.D. and Gelman, A. (2014). The No-U-turn sampler: adaptively setting
         path lengths in Hamiltonian Monte Carlo. Journal of Machine Learning Research,
         15(1), pp.1593-1623.
      2. Betancourt, M. (2013). Generalizing the no-U-turn sampler to Riemannian
         manifolds. arXiv preprint arXiv:1304.1920.
    """
    return (
        np.sum(system.dh_dmom(state_1) * sum_mom) < 0
        or np.sum(system.dh_dmom(state_2) * sum_mom) < 0
    )


class _SubTree(NamedTuple):
    """Sub-tree of binary trajectory tree for dynamic integration transitions."""

    negative: ChainState
    positive: ChainState
    sum_mom: ArrayLike
    weight: ScalarLike
    depth: int


class DynamicIntegrationTransition(IntegrationTransition):
    """Base class for dynamic integration transitions.

    In each transition a binary tree of states is recursively computed by integrating
    randomly forward and backward in time by a number of steps equal to the previous
    tree size until a termination criteria on the tree's subtrees is met. The next chain
    state is chosen from the candidate states using a progressive sampling scheme based
    on relative weights of the different candidate states, with the sampling biased
    towards states further from the current state.

    References:
      1. Hoffman, M.D. and Gelman, A., 2014. The No-U-turn sampler: adaptively setting
         path lengths in Hamiltonian Monte Carlo. Journal of Machine Learning Research,
         15(1), pp.1593-1623.
      2. Betancourt, M., 2017. A conceptual introduction to Hamiltonian Monte Carlo.
         arXiv preprint arXiv:1701.02434.
    """

    def __init__(
        self,
        system: System,
        integrator: Integrator,
        *,
        max_tree_depth: int = 10,
        max_delta_h: float = 1000.0,
        termination_criterion: TerminationCriterion = riemannian_no_u_turn_criterion,
        do_extra_subtree_checks: bool = True,
    ) -> None:
        """
        Args:
            system: Hamiltonian system to be simulated.
            integrator: Symplectic integrator appropriate to the specified Hamiltonian
                system.
            max_tree_depth: Maximum depth to expand trajectory binary tree to. The
                maximum number of integrator steps corresponds to
                :code:`2**max_tree_depth`.
            max_delta_h: Maximum change to tolerate in the Hamiltonian function over a
                trajectory before signalling a divergence.
            termination_criterion: Function computing criterion to use to determine when
                to terminate trajectory tree expansion. The function should take a
                Hamiltonian system as its first argument, a pair of states corresponding
                to the two edge nodes in the trajectory (sub-)tree being checked and an
                array containing the sum of the momentums over the trajectory
                (sub)-tree. Defaults to :py:class:`riemannian_no_u_turn_criterion`.
            do_extra_subtree_checks: Whether to perform additional termination criterion
                checks on overlapping subtrees of the current tree to improve robustness
                in systems with dynamics which are well approximated by independent
                system of simple harmonic oscillators. In such systems (corresponding to
                e.g. a standard normal target distribution and identity metric matrix
                representation) at certain step sizes a 'resonant' behaviour is seen by
                which the termination criterion fails to detect that the trajectory has
                expanded past a half-period i.e. has 'U-turned' resulting in
                trajectories continuing to expand, potentially up until the
                :code:`max_tree_depth` limit is hit. For more details see `this Stan
                Discourse discussion <kutt.it/yAkIES>`_. If
                :code:`do_extra_subtree_checks` is set to :code:`True` additional
                termination criterion checks are performed on overlapping subtrees which
                help to reduce this resonant behaviour at the cost of more conservative
                trajectory termination in some correlated models and some overhead from
                additional checks.
        """
        super().__init__(system, integrator)
        if max_tree_depth <= 0:
            msg = "max_tree_depth must be non-negative."
            raise ValueError(msg)
        self.max_tree_depth = max_tree_depth
        self.max_delta_h = max_delta_h
        self.termination_criterion = termination_criterion
        self.do_extra_subtree_checks = do_extra_subtree_checks
        self._statistic_types["av_metrop_accept_prob"] = (np.float64, np.nan)
        self._statistic_types["reject_prob"] = (np.float64, np.nan)
        self._statistic_types["tree_depth"] = (np.int64, -1)
        self._statistic_types["diverging"] = (bool, False)

    def _termination_criterion(
        self,
        tree: _SubTree,
        neg_subtree: _SubTree,
        pos_subtree: _SubTree,
    ) -> bool:
        # If performing extra subtree checks evaluate lazily i.e. only evaluate
        # if initial whole tree check fails. Extra subtree checks also only
        # performed for trees of depth 2 and above (i.e. containing at least
        # 4 states) as for trees of depth 1 they are redundant.
        if self.termination_criterion(
            self.system,
            tree.negative,
            tree.positive,
            tree.sum_mom,
        ):
            return True
        if tree.depth > 1 and self.do_extra_subtree_checks:
            return self.termination_criterion(
                self.system,
                neg_subtree.negative,
                pos_subtree.negative,
                neg_subtree.sum_mom + pos_subtree.negative.mom,
            ) or self.termination_criterion(
                self.system,
                neg_subtree.positive,
                pos_subtree.positive,
                pos_subtree.sum_mom + neg_subtree.positive.mom,
            )
        return False

    def _new_leave(
        self,
        state: ChainState,
        h: ScalarLike,
        aux_vars: dict[str, ArrayLike],
    ) -> _SubTree:
        return _SubTree(
            negative=state,
            positive=state,
            sum_mom=np.asarray(state.mom),
            weight=self._weight_function(h, aux_vars),
            depth=0,
        )

    def _merge_subtrees(self, neg_subtree: _SubTree, pos_subtree: _SubTree) -> _SubTree:
        if neg_subtree.depth != pos_subtree.depth:
            msg = "Cannot merge subtrees of different depths."
            raise ValueError(msg)
        return _SubTree(
            negative=neg_subtree.negative,
            positive=pos_subtree.positive,
            weight=neg_subtree.weight + pos_subtree.weight,
            sum_mom=neg_subtree.sum_mom + pos_subtree.sum_mom,
            depth=neg_subtree.depth + 1,
        )

    def _init_aux_vars(
        self,
        state: ChainState,
        rng: Generator,  # noqa: ARG002
    ) -> dict[str, ScalarLike]:
        return {"h_init": self.system.h(state)}

    @abstractmethod
    def _weight_function(
        self,
        h: ScalarLike,
        aux_vars: dict[str, ScalarLike],
    ) -> ScalarLike:
        pass

    @abstractmethod
    def _weight_ratio(
        self,
        numerator: ScalarLike,
        denominator: ScalarLike,
    ) -> ScalarLike:
        pass

    @abstractmethod
    def _check_divergence(self, h: float, aux_vars: dict[str, ScalarLike]) -> None:
        pass

    def _build_tree(
        self,
        depth: int,
        state: ChainState,
        stats: dict[str, ScalarLike],
        rng: Generator,
        aux_vars: dict[str, ScalarLike],
    ) -> tuple[bool, _SubTree | None, ChainState | None]:
        if depth == 0:
            # recursion base case
            try:
                # integrate forward/backward one step depending on state.dir
                state = self.integrator.step(state)
                h = self.system.h(state)
                h = np.inf if np.isnan(h) else h
                tree = self._new_leave(state, h, aux_vars)
                proposal = state
                h_diff = aux_vars["h_init"] - h
                # Explicitly check if h_diff is NaN as min(0, NaN) = 0
                metrop_accept_prob = 0.0 if np.isnan(h_diff) else np.exp(min(0, h_diff))
                stats["sum_metrop_accept_prob"] += metrop_accept_prob
                stats["n_step"] += 1
                # default to assuming valid and then check for divergence
                terminate = False
                self._check_divergence(h, aux_vars)
            except IntegratorError as e:
                _process_integrator_error(e, stats)
                terminate, tree, proposal = True, None, None
            return terminate, tree, proposal
        # build 'inner' subtree, i.e. starting from current state
        terminate, inner_tree, inner_proposal = self._build_tree(
            depth - 1,
            state,
            stats,
            rng,
            aux_vars,
        )
        if terminate:
            return terminate, None, None
        # build 'outer' subtree, i.e. starting from terminus of inner subtree
        state = inner_tree.positive if state.dir == 1 else inner_tree.negative
        terminate, outer_tree, outer_proposal = self._build_tree(
            depth - 1,
            state,
            stats,
            rng,
            aux_vars,
        )
        if terminate:
            return terminate, None, None
        # merge two subtrees accounting for integration direction
        neg_subtree = inner_tree if state.dir == 1 else outer_tree
        pos_subtree = outer_tree if state.dir == 1 else inner_tree
        tree = self._merge_subtrees(neg_subtree, pos_subtree)
        # sample new proposal from two subtree proposals according to weights
        accept_outer_prob = self._weight_ratio(outer_tree.weight, tree.weight)
        proposal = (
            outer_proposal if rng.uniform() < accept_outer_prob else inner_proposal
        )
        # check termination criterion on tree and subtrees
        terminate = self._termination_criterion(tree, neg_subtree, pos_subtree)
        return terminate, tree, proposal

    def sample(
        self,
        state: ChainState,
        rng: Generator,
    ) -> tuple[ChainState, dict[str, ScalarLike]]:
        stats = {
            "n_step": 0,
            "sum_metrop_accept_prob": 0.0,
            "reject_prob": 1.0,
            "diverging": False,
            "convergence_error": False,
            "non_reversible_step": False,
            "step_size": self.integrator.step_size,
        }
        aux_vars = self._init_aux_vars(state, rng)
        tree = self._new_leave(state, aux_vars["h_init"], aux_vars)
        next_state = state
        for depth in range(self.max_tree_depth):
            # uniformly sample direction to expand tree in
            direction = 2 * (rng.uniform() < 0.5) - 1  # noqa: PLR2004
            state = tree.positive if direction == 1 else tree.negative
            state.dir = direction
            # expand tree by building new subtree of current depth
            terminate, new_tree, new_proposal = self._build_tree(
                depth,
                state,
                stats,
                rng,
                aux_vars,
            )
            if terminate:
                break
            # progressively sample new state by choosing between
            # current new state and proposal from new subtree, biasing
            # towards the new subtree proposal
            accept_proposal_prob = self._weight_ratio(new_tree.weight, tree.weight)
            if rng.uniform() < accept_proposal_prob:
                next_state = new_proposal
            # each proposal acceptance independent therefore overall probability
            # of 'rejecting' - i.e. not accepting all proposals is product of
            # probabilties of not accepting each proposal
            stats["reject_prob"] *= 1.0 - accept_proposal_prob
            # merge new subtree into current tree accounting for direction
            neg_subtree = tree if direction == 1 else new_tree
            pos_subtree = new_tree if direction == 1 else tree
            tree = self._merge_subtrees(neg_subtree, pos_subtree)
            # check termination criterion on new tree and subtrees
            if self._termination_criterion(tree, neg_subtree, pos_subtree):
                break
        sum_accept_prob = stats.pop("sum_metrop_accept_prob")
        if stats["n_step"] > 0:
            stats["av_metrop_accept_prob"] = sum_accept_prob / stats["n_step"]
        else:
            stats["av_metrop_accept_prob"] = 0.0
        if any(
            stats[key]
            for key in ["diverging", "convergence_error", "non_reversible_step"]
        ):
            stats["accept_stat"] = 0.0
        else:
            stats["accept_stat"] = stats["av_metrop_accept_prob"]
        stats["tree_depth"] = depth
        return next_state, stats


class MultinomialDynamicIntegrationTransition(DynamicIntegrationTransition):
    """Dynamic integration transition with multinomial sampling of new state.

    In each transition a binary tree of states is recursively computed by integrating
    randomly forward and backward in time by a number of steps equal to the previous
    tree size (Hoffman and Gelman, 2014; Betancourt, 2017) until a termination criteria
    on the tree leaves is met. The next chain state is chosen from the candidate states
    using a progressive multinomial sampling scheme (Betancourt, 2017) based on the
    relative probability densities of the different candidate states, with the sampling
    biased towards states further from the current state.

    References:
      1. Hoffman, M.D. and Gelman, A. (2014). The No-U-turn sampler: adaptively setting
         path lengths in Hamiltonian Monte Carlo. Journal of Machine Learning Research,
         15(1), pp.1593-1623.
      2. Betancourt, M. (2017). A conceptual introduction to Hamiltonian Monte Carlo.
         arXiv preprint arXiv:1701.02434.
    """

    def _weight_function(
        self,
        h: ScalarLike,
        aux_vars: dict[str, ScalarLike],  # noqa: ARG002
    ) -> ScalarLike:
        return LogRepFloat(log_val=-h)

    def _weight_ratio(
        self,
        numerator: ScalarLike,
        denominator: ScalarLike,
    ) -> ScalarLike:
        return min(numerator / denominator, 1)

    def _check_divergence(self, h: ScalarLike, aux_vars: dict[str, ScalarLike]) -> None:
        if h - aux_vars["h_init"] > self.max_delta_h:
            msg = f"delta_h = {h - aux_vars['h_init']}"
            raise HamiltonianDivergenceError(msg)


class SliceDynamicIntegrationTransition(DynamicIntegrationTransition):
    """Dynamic integration transition with slice sampling of new state.

    In each transition a binary tree of states is recursively computed by integrating
    randomly forward and backward in time by a number of steps equal to the previous
    tree size until a termination criteria on the tree leaves is met. The next chain
    state is chosen from the candidate states using a progressive slice sampling scheme
    based on the relative probability densities of the different candidate states, with
    the slice sampler biased towards states further from the current state.

    When used with the :py:class:`euclidean_no_u_turn_criterion` this transition is
    equivalent to the transitions in 'Algorithm 3: Efficient No-U-Turn Sampler' in
    Hoffman and Gelman (2014).

    References:
      1. Hoffman, M.D. and Gelman, A. (2014). The No-U-turn sampler: adaptively setting
         path lengths in Hamiltonian Monte Carlo. Journal of Machine Learning Research,
         15(1), pp.1593-1623.
    """

    def _init_aux_vars(
        self,
        state: ChainState,
        rng: Generator,
    ) -> dict[str, ScalarLike]:
        aux_vars = super()._init_aux_vars(state, rng)
        aux_vars["log_u"] = np.log(rng.uniform()) - aux_vars["h_init"]
        return aux_vars

    def _weight_function(
        self,
        h: ScalarLike,
        aux_vars: dict[str, ScalarLike],
    ) -> ScalarLike:
        return (aux_vars["log_u"] <= -h) * 1

    def _weight_ratio(
        self,
        numerator: ScalarLike,
        denominator: ScalarLike,
    ) -> ScalarLike:
        return min(numerator / denominator, 1) if denominator > 0 else min(numerator, 1)

    def _check_divergence(self, h: ScalarLike, aux_vars: dict[str, ScalarLike]) -> None:
        if h + aux_vars["log_u"] > self.max_delta_h:
            msg = f"delta_h = {h + aux_vars['log_u']}"
            raise HamiltonianDivergenceError(msg)



"""Solvers for non-linear systems of equations for implicit integrators."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

import numpy as np

from mici.errors import ConvergenceError, LinAlgError

if TYPE_CHECKING:
    from mici.states import ChainState
    from mici.systems import (
        ConstrainedEuclideanMetricSystem,
        ConstrainedTractableFlowSystem,
    )
    from mici.types import ArrayFunction, ArrayLike, ScalarFunction, ScalarLike


def euclidean_norm(vct: ArrayLike) -> ScalarLike:
    """Calculate the Euclidean (L-2) norm of a vector."""
    return (vct**2).sum() ** 0.5


def maximum_norm(vct: ArrayLike) -> ScalarLike:
    """Calculate the maximum (L-infinity) norm of a vector."""
    return (abs(vct)).max()


class FixedPointSolver(Protocol):
    """Solver for fixed point equation :code:`func(x) = x`."""

    def __call__(
        self, func: ArrayFunction, x0: ArrayLike, **kwargs: float | ScalarFunction
    ) -> ArrayLike:
        """Solve fixed point equation.

        Args:
            func: Function to solve for fixed point of.
            x0: Point to initialize solver at.

        Returns:
            Fixed point solved for.
        """


def solve_fixed_point_direct(
    func: ArrayFunction,
    x0: ArrayLike,
    convergence_tol: float = 1e-9,
    divergence_tol: float = 1e10,
    max_iters: int = 100,
    norm: ScalarFunction = maximum_norm,
) -> ArrayLike:
    """Solve fixed point equation :code:`func(x) = x` using direct iteration.

    Args:
        func: Function to find fixed point of.
        x0: Initial state (function argument).
        convergence_tol: Convergence tolerance - solver successfully terminates when
            :code:`norm(func(x) - x) < convergence_tol`.
        divergence_tol: Divergence tolerance - solver aborts if
            :code:`norm(func(x) - x) > divergence_tol` on any iteration.
        max_iters: Maximum number of iterations before raising exception.
        norm: Norm to use to assess convergence.

    Returns:
        Solution to fixed point equation with
        :code:`norm(func(x) - x) < convergence_tol`.

    Raises:
        mici.errors.ConvergenceError: If solver does not converge within
            :code:`max_iters` iterations, diverges or encounters a :py:exc:`ValueError`
            during the iteration.
    """
    try:
        for i in range(max_iters):
            x = func(x0)
            error = norm(x - x0)
            if error > divergence_tol or np.isnan(error):
                msg = (
                    f"Fixed point iteration diverged on iteration {i}. "
                    f"Last error={error:.1e}."
                )
                raise ConvergenceError(msg)
            if error < convergence_tol:
                return x
            x0 = x
    except (ValueError, LinAlgError) as e:
        # Make robust to errors in intermediate linear algebra ops
        msg = f"{type(e)} at iteration {i} of fixed point solver ({e})."
        raise ConvergenceError(msg) from e
    msg = f"Fixed point iteration did not converge. Last error={error:.1e}."
    raise ConvergenceError(msg)


def solve_fixed_point_steffensen(
    func: ArrayFunction,
    x0: ArrayLike,
    convergence_tol: float = 1e-9,
    divergence_tol: float = 1e10,
    max_iters: int = 100,
    norm: ScalarFunction = maximum_norm,
) -> ArrayLike:
    """Solve fixed point equation :code:`func(x) = x` using Steffensen's method.

    Steffennsen's method achieves quadratic convergence but at the cost of two function
    evaluations per iteration so for functions where convergence is achieved in a small
    number of iterations, direct iteration may be cheaper.

    Args:
        func: Function to find fixed point of.
        x0: Initial state (function argument).
        convergence_tol: Convergence tolerance - solver successfully terminates
            when :code:`norm(func(x) - x) < convergence_tol`.
        divergence_tol: Divergence tolerance - solver aborts if
            :code:`norm(func(x) - x) > divergence_tol` on any iteration.
        max_iters: Maximum number of iterations before raising exception.
        norm: Norm to use to assess convergence.

    Returns:
        Solution to fixed point equation with
        :code:`norm(func(x) - x) < convergence_tol`.

    Raises:
        mici.errors.ConvergenceError: If solver does not converge within
            :code:`max_iters` iterations, diverges or encounters a :py:exc:`ValueError`
            during the iteration.
    """
    try:
        for i in range(max_iters):
            x1 = func(x0)
            x2 = func(x1)
            denom = x2 - 2 * x1 + x0
            # Set any zero values in denominator of update term to smalllest
            # floating point value to prevent divide-by-zero errors
            denom[abs(denom) == 0.0] = np.finfo(x0.dtype).eps
            x = x0 - (x1 - x0) ** 2 / denom
            error = norm(x - x0)
            if error > divergence_tol or np.isnan(error):
                msg = (
                    f"Fixed point iteration diverged on iteration {i}. "
                    f"Last error={error:.1e}."
                )
                raise ConvergenceError(msg)
            if error < convergence_tol:
                return x
            x0 = x
    except (ValueError, LinAlgError) as e:
        # Make robust to errors in intermediate linear algebra ops
        msg = f"{type(e)} at iteration {i} of fixed point solver ({e})."
        raise ConvergenceError(msg) from e
    msg = f"Fixed point iteration did not converge. Last error={error:.1e}."
    raise ConvergenceError(msg)


class ProjectionSolver(Protocol):
    r"""Solver for projection on to manifold step in constrained integrator.

    Solves an equation of the form

    .. math::

        r(\lambda) = c(\Phi_{2,1}(t)(q, p + \partial c(q)^T \lambda)) = 0,

    for the vector of Lagrange multipliers :math:`\lambda` to project a point on to the
    manifold defined by the zero level set of a constraint function :math:`c`, with
    :math:`\Phi_{2,1}` the flow map for the :math:`h_2` Hamiltonian component for the
    system restricted to the position component output. The map :math:`\Phi_{2,1}` is
    assumed to be linear in its second (momentum) argument.
    """

    def __call__(
        self,
        state: ChainState,
        state_prev: ChainState,
        time_step: float,
        system: ConstrainedTractableFlowSystem,
        **kwargs: float | ScalarFunction,
    ) -> ChainState:
        """Solve for projection on to manifold step.

        Args:
            state: Current chain state after unconstrained step.
            state_prev: Previous chain state on manifold.
            time_step: Integrator time step for unconstrained step.
            system: Hamiltonian system constrained dynamics are being simulated for.

        Returns:
            Chain state after projection on to manifold.
        """
        ...


def solve_projection_onto_manifold_quasi_newton(
    state: ChainState,
    state_prev: ChainState,
    time_step: float,
    system: ConstrainedEuclideanMetricSystem,
    constraint_tol: float = 1e-9,
    position_tol: float = 1e-8,
    divergence_tol: float = 1e10,
    max_iters: int = 50,
    norm: ScalarFunction = maximum_norm,
) -> ChainState:
    r"""Solve constraint equation using symmetric quasi-Newton method.

    Only requires re-evaluating the constraint function :code:`system.constr` within the
    solver loop and no recomputation of matrix decompositions on each iteration.

    Solves an equation of the form

    .. math::

        r(\lambda) = c(\Phi_{2,1}(t)(q, p + \partial c(q)^T \lambda)) = 0,

    for the vector of Lagrange multipliers :math:`\lambda` to project a point on to the
    manifold defined by the zero level set of a constraint function :math:`c`, with
    :math:`\Phi_{2,1}` the flow map for the :math:`h_2` Hamiltonian component for the
    system restricted to the position component output.

    The Jacobian of the residual function :math:`r` is

    .. math::

        \partial r(\lambda) =
        \partial c(\Phi_{2,1}(t)(q, p + \partial c(q)^T \lambda))
        \partial_2 (\Phi_{2,1}(t))(q, p + \partial c(q)^T \lambda)
        \partial c(q)^T

    where :math:`\partial_2 (\Phi_{2,1}(t))` is the Jacobian of the (position
    restricted) flow-map :math:`\Phi_{2,1}` with respect to its second (momentum)
    argument. We assume here that :math:`\Phi_{2,1}(t)` is linear in its second
    (momentum) argument, such that :math:`\partial_2 (\Phi_{2,1}(t))(q, p + \partial
    c(q)^T \lambda)` is constant with respect to :math:`\lambda`, and henceforth use the
    shorthand :math:`\partial_2 (\Phi_{2,1}(t))` to refer to this matrix.

    The full Newton update is

    .. math::

        \lambda' = \lambda - \partial r(\lambda)^{-1} r(\lambda)

    which requires evaluating :math:`\partial c` on each iteration and solving a linear
    system in the residual Jacobian :math:`\partial r(\lambda)`.

    The symmetric quasi-Newton iteration instead uses the approximation

    .. math::

        \partial c(\Phi_{2,1}(t)(q, p + \partial c(q)^T\lambda))
        \partial_2 (\Phi_{2,1}(t))
        \partial c(q)^T
        \approx
        \partial c(q)
        \partial_2 (\Phi_{2,1}(t))
        \partial c(q)^T

    with the corresponding update

    .. math::

        \lambda' =
        \lambda
        - (\partial c(q) \partial_2 (\Phi_{2,1}(t)) \partial c(q)^T)^{-1} r(\lambda)

    allowing a previously computed decomposition of the matrix

    .. math::

        \partial c(q) \partial_2 (\Phi_{2,1}(t)) \partial c(q)^T,

    to be used to solve the linear system in each iteration with no requirement to
    evaluate :math:`\partial c` (:code:`system.jacob_constr`) on each iteration.

    Args:
        state: Post :code:`h2_flow` update state to project.
        state_prev: Previous state in co-tangent bundle before :code:`h2_flow` update
            which defines the co-tangent space to perform  projection in.
        time_step: Integrator time step used in :code:`h2_flow` update.
        system: Hamiltonian system defining :code:`h2_flow` and :code:`constr` functions
            used to define constraint equation to solve.
        constraint_tol: Convergence tolerance in constraint space. Iteration will
            continue until :code:`norm(constr(pos)) < constraint_tol` where :code:`pos`
            is the position at the current iteration.
        position_tol: Convergence tolerance in position space. Iteration will continue
            until :code:`norm(delta_pos) < position_tol` where :code:`delta_pos` is the
            change in the position in the current iteration.
        divergence_tol: Divergence tolerance - solver aborts if
            :code:`norm(constr(pos)) > divergence_tol` on any iteration where
            :code:`pos` is the position at the current iteration and raises
            :py:exc:`mici.errors.ConvergenceError`.
        max_iters: Maximum number of iterations to perform before aborting and raising
            :py:exc:`mici.errors.ConvergenceError`.
        norm: Norm to use to test for convergence.

    Returns:
        Updated state object with position component satisfying constraint equation to
        within :code:`constraint_tol`, i.e.
        :code:`norm(system.constr(state.pos)) < constraint_tol`.

    Raises:
        mici.errors.ConvergenceError: If solver does not converge within
            :code:`max_iters` iterations, diverges or encounters a :py:exc:`ValueError`
            during the iteration.
    """
    mu = np.zeros_like(state.pos)
    jacob_constr_prev = system.jacob_constr(state_prev)
    # Use absolute value of dt and adjust for sign of dt in mom update below
    dh2_flow_pos_dmom, dh2_flow_mom_dmom = system.dh2_flow_dmom(
        state_prev,
        abs(time_step),
    )
    inv_jacob_constr_inner_product = system.jacob_constr_inner_product(
        jacob_constr_prev,
        dh2_flow_pos_dmom,
    ).inv
    try:
        for i in range(max_iters):
            constr = system.constr(state)
            error = norm(constr)
            delta_mu = jacob_constr_prev.T @ (inv_jacob_constr_inner_product @ constr)
            delta_pos = dh2_flow_pos_dmom @ delta_mu
            if error > divergence_tol or np.isnan(error):
                msg = (
                    f"Quasi-Newton solver diverged on iteration {i}. "
                    f"Last |constr|={error:.1e}, |delta_pos|={norm(delta_pos):.1e}."
                )
                raise ConvergenceError(msg)
            if error < constraint_tol and norm(delta_pos) < position_tol:
                state.mom -= np.sign(time_step) * dh2_flow_mom_dmom @ mu
                return state
            mu += delta_mu
            state.pos -= delta_pos
    except (ValueError, LinAlgError) as e:
        # Make robust to errors in intermediate linear algebra ops
        msg = f"{type(e)} at iteration {i} of quasi-Newton solver ({e})."
        raise ConvergenceError(msg) from e
    msg = (
        f"Quasi-Newton solver did not converge with {max_iters} iterations. "
        f"Last |constr|={error:.1e}, |delta_pos|={norm(delta_pos)}."
    )
    raise ConvergenceError(msg)


def solve_projection_onto_manifold_newton(
    state: ChainState,
    state_prev: ChainState,
    time_step: float,
    system: ConstrainedEuclideanMetricSystem,
    constraint_tol: float = 1e-9,
    position_tol: float = 1e-8,
    divergence_tol: float = 1e10,
    max_iters: int = 50,
    norm: ScalarFunction = maximum_norm,
) -> ChainState:
    r"""Solve constraint equation using Newton's method.

    Requires re-evaluating both the constraint function :code:`system.constr` and
    constraint Jacobian :code:`system.jacob_constr` within the solver loop and
    computation of matrix decompositions on each iteration.

    Solves an equation of the form

    .. math::

        r(\lambda) = c(\Phi_{2,1}(t)(q, p + \partial c(q)^T \lambda)) = 0,

    for the vector of Lagrange multipliers :math:`\lambda` to project a point on to the
    manifold defined by the zero level set of a constraint function :math:`c`, with
    :math:`\Phi_{2,1}` the flow map for the :math:`h_2` Hamiltonian component for the
    system restricted to the position component output.

    The Jacobian of the residual function :math:`r` is

    .. math::

        \partial r(\lambda) =
        \partial c(\Phi_{2,1}(t)(q, p + \partial c(q)^T \lambda))
        \partial_2 (\Phi_{2,1}(t))(q, p + \partial c(q)^T \lambda)
        \partial c(q)^T

    where :math:`\partial_2 (\Phi_{2,1}(t))` is the Jacobian of the (position
    restricted) flow-map :math:`\Phi_{2,1}` with respect to its second (momentum)
    argument. We assume here that :math:`\Phi_{2,1}(t)` is linear in its second
    (momentum) argument, such that :math:`\partial_2 (\Phi_{2,1}(t))(q, p + \partial
    c(q)^T \lambda)` is constant with respect to :math:`\lambda`.

    The full Newton update is

    .. math::

        \lambda' = \lambda - \partial r(\lambda)^{-1} r(\lambda)

    which requires evaluating :math:`\partial c` on each iteration and solving a linear
    system in the residual Jacobian :math:`\partial r(\lambda)`.

    Args:
        state: Post :code:`h2_flow` update state to project.
        state_prev: Previous state in co-tangent bundle before :code:`h2_flow` update
            which defines the co-tangent space to perform  projection in.
        time_step: Integrator time step used in :code:`h2_flow` update.
        system: Hamiltonian system defining :code:`h2_flow` and :code:`constr` functions
            used to define constraint equation to solve.
        constraint_tol: Convergence tolerance in constraint space. Iteration will
            continue until :code:`norm(constr(pos)) < constraint_tol` where :code:`pos`
            is the position at the current iteration.
        position_tol: Convergence tolerance in position space. Iteration will continue
            until :code:`norm(delta_pos) < position_tol` where :code:`delta_pos` is the
            change in the position in the current iteration.
        divergence_tol: Divergence tolerance - solver aborts if
            :code:`norm(constr(pos)) > divergence_tol` on any iteration where
            :code:`pos` is the position at the current iteration and raises
            :py:exc:`mici.errors.ConvergenceError`.
        max_iters: Maximum number of iterations to perform before aborting and raising
            :py:exc:`mici.errors.ConvergenceError`.
        norm: Norm to use to test for convergence.

    Returns:
        Updated state object with position component satisfying constraint equation to
        within :code:`constraint_tol`, i.e.
        :code:`norm(system.constr(state.pos)) < constraint_tol`.

    Raises:
        mici.errors.ConvergenceError: If solver does not converge within
            :code:`max_iters` iterations, diverges or encounters a :py:exc:`ValueError`
            during the iteration.
    """
    mu = np.zeros_like(state.pos)
    jacob_constr_prev = system.jacob_constr(state_prev)
    # Use absolute value of dt and adjust for sign of dt in mom update below
    dh2_flow_pos_dmom, dh2_flow_mom_dmom = system.dh2_flow_dmom(
        state_prev,
        abs(time_step),
    )
    try:
        for i in range(max_iters):
            jacob_constr = system.jacob_constr(state)
            constr = system.constr(state)
            error = norm(constr)
            delta_mu = jacob_constr_prev.T @ (
                system.jacob_constr_inner_product(
                    jacob_constr,
                    dh2_flow_pos_dmom,
                    jacob_constr_prev,
                ).inv
                @ constr
            )
            delta_pos = dh2_flow_pos_dmom @ delta_mu
            if error > divergence_tol or np.isnan(error):
                msg = (
                    f"Newton solver diverged at iteration {i}. "
                    f"Last |constr|={error:.1e}, |delta_pos|={norm(delta_pos):.1e}."
                )
                raise ConvergenceError(msg)
            if error < constraint_tol and norm(delta_pos) < position_tol:
                state.mom -= np.sign(time_step) * dh2_flow_mom_dmom @ mu
                return state
            mu += delta_mu
            state.pos -= delta_pos
    except (ValueError, LinAlgError) as e:
        # Make robust to errors in intermediate linear algebra ops
        msg = f"{type(e)} at iteration {i} of Newton solver ({e})."
        raise ConvergenceError(msg) from e
    msg = (
        f"Newton solver did not converge in {max_iters} iterations. "
        f"Last |constr|={error:.1e}, |delta_pos|={norm(delta_pos)}."
    )
    raise ConvergenceError(msg)


def solve_projection_onto_manifold_newton_with_line_search(
    state: ChainState,
    state_prev: ChainState,
    time_step: float,
    system: ConstrainedEuclideanMetricSystem,
    constraint_tol: float = 1e-9,
    position_tol: float = 1e-8,
    divergence_tol: float = 1e10,
    max_iters: int = 50,
    max_line_search_iters: int = 10,
    norm: ScalarFunction = maximum_norm,
) -> ChainState:
    r"""Solve constraint equation using Newton's method with backtracking line-search.

    Requires re-evaluating both the constraint function :code:`system.constr` and
    constraint Jacobian :code:`system.jacob_constr` within the solver loop and
    computation of matrix decompositions on each iteration.

    Solves an equation of the form

    .. math::

        r(\lambda) = c(\Phi_{2,1}(t)(q, p + \partial c(q)^T \lambda)) = 0,

    for the vector of Lagrange multipliers :math:`\lambda` to project a point on to the
    manifold defined by the zero level set of a constraint function :math:`c`, with
    :math:`\Phi_{2,1}` the flow map for the :math:`h_2` Hamiltonian component for the
    system restricted to the position component output.

    The Jacobian of the residual function :math:`r` is

    .. math::

        \partial r(\lambda) =
        \partial c(\Phi_{2,1}(t)(q, p + \partial c(q)^T \lambda))
        \partial_2 (\Phi_{2,1}(t))(q, p + \partial c(q)^T \lambda)
        \partial c(q)^T

    where :math:`\partial_2 (\Phi_{2,1}(t))` is the Jacobian of the (position
    restricted) flow-map :math:`\Phi_{2,1}` with respect to its second (momentum)
    argument. We assume here that :math:`\Phi_{2,1}(t)` is linear in its second
    (momentum) argument, such that :math:`\partial_2 (\Phi_{2,1}(t))(q, p + \partial
    c(q)^T \lambda)` is constant with respect to :math:`\lambda`.

    The scaled Newton update is

    .. math::

        \lambda'(\alpha) = \lambda - \alpha \partial r(\lambda)^{-1} r(\lambda)

    which requires evaluating :math:`\partial c` on each iteration and solving a linear
    system in the residual Jacobian :math:`\partial r(\lambda)`.

    The step size :math:`\alpha \in [0, 1]` is initialised at :math:`\alpha = 1` with a
    backtracking line search performed by multiplying :math:`\alpha` by 0.5 until
    :math:`|r(\lambda'(\alpha))| < |r(\lambda)`.

    Args:
        state: Post :code:`h2_flow` update state to project.
        state_prev: Previous state in co-tangent bundle before :code:`h2_flow` update
            which defines the co-tangent space to perform  projection in.
        time_step: Integrator time step used in :code:`h2_flow` update.
        system: Hamiltonian system defining :code:`h2_flow` and :code:`constr` functions
            used to define constraint equation to solve.
        constraint_tol: Convergence tolerance in constraint space. Iteration will
            continue until :code:`norm(constr(pos)) < constraint_tol` where :code:`pos`
            is the position at the current iteration.
        position_tol: Convergence tolerance in position space. Iteration will continue
            until :code:`norm(delta_pos) < position_tol` where :code:`delta_pos` is the
            change in the position in the current iteration.
        divergence_tol: Divergence tolerance - solver aborts if
            :code:`norm(constr(pos)) > divergence_tol` on any iteration where
            :code:`pos` is the position at the current iteration and raises
            :py:exc:`mici.errors.ConvergenceError`.
        max_iters: Maximum number of iterations to perform before aborting and raising
            :py:exc:`mici.errors.ConvergenceError`.
        max_line_search_iters: Maximum number of 'inner' line search iterations to
            perform to try to find step size along search direction which decreases
            residual norm.
        norm: Norm to use to test for convergence.

    Returns:
        Updated state object with position component satisfying constraint equation to
        within :code:`constraint_tol`, i.e.
        :code:`norm(system.constr(state.pos)) < constraint_tol`.

    Raises:
        mici.errors.ConvergenceError: If solver does not converge within
            :code:`max_iters` iterations, diverges or encounters a :py:exc:`ValueError`
            during the iteration.
    """
    mu = np.zeros_like(state.pos)
    jacob_constr_prev = system.jacob_constr(state_prev)
    # Use absolute value of dt and adjust for sign of dt in mom update below
    dh2_flow_pos_dmom, dh2_flow_mom_dmom = system.dh2_flow_dmom(
        state_prev,
        abs(time_step),
    )
    # Initialize with dummy values to avoid undefined name linter errors
    delta_pos, step_size = None, None
    for i in range(max_iters):
        try:
            jacob_constr = system.jacob_constr(state)
            constr = system.constr(state)
            error = norm(constr)
            if i > 0 and (error > divergence_tol or np.isnan(error)):
                msg = (
                    f"Newton solver diverged at iteration {i}. "
                    f"Last |constr|={error:.1e}, |delta_pos|={norm(delta_pos):.1e}."
                )
                raise ConvergenceError(msg)
            if error < constraint_tol and (
                i == 0 or norm(step_size * delta_pos) < position_tol
            ):
                state.mom -= np.sign(time_step) * dh2_flow_mom_dmom @ mu
                return state
            delta_mu = jacob_constr_prev.T @ (
                system.jacob_constr_inner_product(
                    jacob_constr,
                    dh2_flow_pos_dmom,
                    jacob_constr_prev,
                ).inv
                @ constr
            )
            delta_pos = -dh2_flow_pos_dmom @ delta_mu
            pos_curr = state.pos.copy()
            step_size = 1.0
            for _ in range(max_line_search_iters):
                state.pos = pos_curr + step_size * delta_pos
                new_error = norm(system.constr(state))
                if new_error < error:
                    break
                step_size *= 0.5
            mu += step_size * delta_mu
        except (ValueError, LinAlgError) as e:
            # Make robust to errors in intermediate linear algebra ops
            msg = f"{type(e)} at iteration {i} of Newton solver ({e})."
            raise ConvergenceError(msg) from e
    msg = (
        f"Newton solver did not converge in {max_iters} iterations. "
        f"Last |constr|={error:.1e}, |delta_pos|={norm(step_size * delta_pos)}."
    )
    raise ConvergenceError(msg)



import pytest

import mici

N_MAIN_ITER = 1


class StagerTests:
    def test_stages(self, stager, n_warm_up_iter, adapters):
        stages = stager.stages(
            n_warm_up_iter,
            N_MAIN_ITER,
            adapters,
            [lambda _: None],
        )
        assert isinstance(stages, dict)
        n_total_iter = 0
        for stage_key, stage in stages.items():
            assert isinstance(stage_key, str)
            assert isinstance(stage, mici.stagers.ChainStage)
            assert stage.n_iter >= 0
            assert stage.adapters is None or isinstance(stage.adapters, dict)
            assert stage.trace_funcs is None or isinstance(stage.trace_funcs, tuple)
            n_total_iter += stage.n_iter
        assert n_total_iter == n_warm_up_iter + N_MAIN_ITER


class TestWarmUpStager(StagerTests):
    @pytest.fixture
    def adapters(self):
        return {
            "integration_transition": [mici.adapters.DualAveragingStepSizeAdapter()],
        }

    @pytest.fixture
    def stager(self):
        return mici.stagers.WarmUpStager()

    @pytest.fixture
    def n_warm_up_iter(self):
        return 1


class TestWindowedWarmUpStager(StagerTests):
    @pytest.fixture
    def adapters(self):
        return {
            "integration_transition": [
                mici.adapters.DualAveragingStepSizeAdapter(),
                mici.adapters.OnlineVarianceMetricAdapter(),
            ],
        }

    @pytest.fixture(params=((), (125, 50, 25, 3)))
    def stager(self, request):
        return mici.stagers.WindowedWarmUpStager(*request.param)

    @pytest.fixture(params=(5, 10, 100, 107, 500, 1003))
    def n_warm_up_iter(self, request):
        return request.param



"""JAX differential operators and helper functions."""

from __future__ import annotations

from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING, ParamSpec, TypeAlias, TypeVar

import numpy as np

JAX_AVAILABLE = True
try:
    import jax
    import jax.numpy as jnp
    from jax import Array as JaxArray
except ImportError:
    JAX_AVAILABLE = False
    JaxArray = TypeVar("JaxArray")

if TYPE_CHECKING:
    from mici.types import (
        ArrayFunction,
        ArrayLike,
        GradientFunction,
        HessianFunction,
        JacobianFunction,
        MatrixHessianProduct,
        MatrixHessianProductFunction,
        MatrixTressianProduct,
        MatrixTressianProductFunction,
        ScalarFunction,
        ScalarLike,
        VectorJacobianProduct,
        VectorJacobianProductFunction,
    )


P = ParamSpec("P")
JaxArrayFunction: TypeAlias = Callable[
    P, "JaxArray | JaxArrayFunction | tuple[JaxArray | JaxArrayFunction, ...]"
]
NumPyArrayFunction: TypeAlias = Callable[
    P, "np.ndarray | NumPyArrayFunction | tuple[np.ndarray | NumPyArrayFunction, ...]"
]


def jit_and_return_numpy_arrays(function: JaxArrayFunction) -> NumPyArrayFunction:
    """Wrap a JIT compiled function returning JAX arrays to instead return NumPy arrays.

    Args:
        function: Function to wrap. Should return one of: a single JAX array, a callable
            returning one or more JAX array or a tuple of one or more JAX arrays or
            functions returning one or more JAX arrays.
        **jit_kwargs: Any keyword arguments to pass to `jax.jit` operator.

    Returns:
        Wrapped function. Any values returned by original function which are JAX arrays
        will instead be NumPy arrays, while any values which are callables returning
        JAX arrays will instead return NumPy arrays.
    """
    jitted_function = jax.jit(function)
    return return_numpy_arrays(jitted_function)


def return_numpy_arrays(function: JaxArrayFunction) -> NumPyArrayFunction:
    """Wrap a function returning JAX arrays to instead return NumPy arrays.

    Args:
        function: Function to wrap. Should return one of: a single JAX array, a callable
            returning one or more JAX array or a tuple of one or more JAX arrays or
            functions returning one or more JAX arrays.

    Returns:
        Wrapped function. Any values returned by original function which are JAX arrays
        will instead be NumPy arrays, while any values which are callables returning
        JAX arrays will instead return NumPy arrays.
    """

    def as_numpy_array(
        value: JaxArray | JaxArrayFunction,
    ) -> np.ndarray | NumPyArrayFunction:
        if callable(value):
            return return_numpy_arrays(value)
        if isinstance(value, jax.Array):
            return np.asarray(value)
        return value

    def function_returning_numpy_arrays(
        *args: P.args, **kwargs: P.kwargs
    ) -> np.ndarray | NumPyArrayFunction | tuple[np.ndarray | NumPyArrayFunction, ...]:
        return_value = function(*args, **kwargs)
        if isinstance(return_value, tuple):
            return tuple(as_numpy_array(value) for value in return_value)
        return as_numpy_array(return_value)

    return function_returning_numpy_arrays


def grad_and_value(func: ScalarFunction) -> GradientFunction:
    """Makes a function that returns both the gradient and value of a function."""

    def grad_and_value_func(x: ArrayLike) -> tuple[ArrayLike, ScalarLike]:
        value, grad = jax.value_and_grad(func)(x)
        return grad, value

    return grad_and_value_func


def _detuple_vjp(vjp_func: jax.tree_util.Partial) -> jax.tree_util.Partial:
    """Transform a VJP of function with one return value so it returns an array."""
    return jax.tree_util.Partial(
        lambda *args, **kwargs: vjp_func.func(*args, **kwargs)[0],
        *vjp_func.args,
        **vjp_func.keywords,
    )


def vjp_and_value(func: ArrayFunction) -> VectorJacobianProductFunction:
    """Makes a function that returns vector-Jacobian-product and value of a function.

    For a vector-valued function `fun` the vector-Jacobian-product (VJP) is here
    defined as a function of a vector `v` corresponding to

        vjp(v) = v @ j

    where `j` is the Jacobian of `f = fun(x)` wrt `x` i.e. the rank-2
    tensor of first-order partial derivatives of the vector-valued function,
    such that

        j[i, k] = ∂f[i] / ∂x[k]
    """

    def vjp_and_value_func(x: ArrayLike) -> tuple[VectorJacobianProduct, ArrayLike]:
        value, vjp = jax.vjp(func, x)
        return _detuple_vjp(vjp), value

    return vjp_and_value_func


def jacobian_and_value(func: ArrayFunction) -> JacobianFunction:
    """Makes a function that returns both the Jacobian and value of a function."""

    def jacobian_and_value_func(x: ArrayLike) -> tuple[ArrayLike, ArrayLike]:
        value, pullback = jax.vjp(func, x)
        basis = jnp.eye(value.size, dtype=value.dtype)
        (jac,) = jax.vmap(pullback)(basis)
        return jac, value

    return jacobian_and_value_func


def mhp_jacobian_and_value(func: ArrayFunction) -> MatrixHessianProductFunction:
    """Makes a function that returns MHP, Jacobian and value of a function.

    For a vector-valued function `fun` the matrix-Hessian-product (MHP) is here
    defined as a function of a matrix `m` corresponding to

        mhp(m) = sum(m[:, :, None] * h[:, :, :], axis=(0, 1))

    where `h` is the vector-Hessian of `f = fun(x)` wrt `x` i.e. the rank-3
    tensor of second-order partial derivatives of the vector-valued function,
    such that

        h[i, j, k] = ∂²f[i] / (∂x[j] ∂x[k])
    """

    def mhp_jacobian_and_value_func(
        x: ArrayLike,
    ) -> tuple[MatrixHessianProduct, ArrayLike, ArrayLike]:
        jac, mhp, value = jax.vjp(jacobian_and_value(func), x, has_aux=True)
        return _detuple_vjp(mhp), jac, value

    return mhp_jacobian_and_value_func


def hessian_grad_and_value(func: ScalarFunction) -> HessianFunction:
    """Makes a function that returns the Hessian, gradient and value of a function."""

    def hessian_grad_and_value_func(
        x: ArrayLike,
    ) -> tuple[ArrayLike, ArrayLike, ScalarLike]:
        basis = jnp.eye(x.size, dtype=x.dtype)
        grad_and_value_func = grad_and_value(func)
        pushforward = partial(jax.jvp, grad_and_value_func, (x,), has_aux=True)
        grad, hessian, value = jax.vmap(pushforward, out_axes=(None, -1, None))(
            (basis,),
        )
        return hessian, grad, value

    return hessian_grad_and_value_func


def mtp_hessian_grad_and_value(func: ScalarFunction) -> MatrixTressianProductFunction:
    """Makes a function that returns MTP, Jacobian and value of a function.

    For a scalar-valued function `fun` the matrix-Tressian-product (MTP) is
    here defined as a function of a matrix `m` corresponding to

        mtp(m) = sum(m[:, :] * t[:, :, :], axis=(-1, -2))

    where `t` is the 'Tressian' of `f = fun(x)` wrt `x` i.e. the 3D array of
    third-order partial derivatives of the scalar-valued function such that

        t[i, j, k] = ∂³f / (∂x[i] ∂x[j] ∂x[k])
    """

    def hessian_and_aux_func(
        x: ArrayLike,
    ) -> tuple[ArrayLike, tuple[ArrayLike, ScalarLike]]:
        hessian, grad, value = hessian_grad_and_value(func)(x)
        return hessian, (grad, value)

    def mtp_hessian_grad_and_value_func(
        x: ArrayLike,
    ) -> tuple[MatrixTressianProduct, ArrayLike, ArrayLike, ScalarLike]:
        hessian, mtp, (grad, value) = jax.vjp(hessian_and_aux_func, x, has_aux=True)
        return _detuple_vjp(mtp), hessian, grad, value

    return mtp_hessian_grad_and_value_func
